from __future__ import annotations

import csv
import json
import os
import re
import warnings
from dataclasses import dataclass
from pathlib import Path
from typing import Optional, List, Iterable, Dict, Iterator, TYPE_CHECKING, Any

from pkm.api.dependencies.dependency import Dependency
from pkm.api.packages.package_metadata import PackageMetadata
from pkm.api.versions.version import Version, StandardVersion
from pkm.config.configuration import IniFileConfiguration, FileConfiguration, computed_based_on
from pkm.utils.commons import UnsupportedOperationException
from pkm.utils.dicts import get_or_compute
from pkm.utils.entrypoints import EntryPoint, ObjectReference
from pkm.utils.files import path_to, resolve_relativity
from pkm.utils.hashes import HashSignature
from pkm.utils.iterators import groupby
from pkm.utils.properties import cached_property

if TYPE_CHECKING:
    from pkm.api.packages.package import PackageDescriptor


class DistInfo:

    def __init__(self, path: Path):
        self.path = path

    @cached_property
    def package_name(self):
        """
        :return: the package name as can be computed from the dist-info file name. note that if the loaded dist-info
         path has non-standard, the return value of this method is undetermined
        """
        return self.path.name.split("-")[0]

    def load_wheel_cfg(self) -> "WheelFileConfiguration":
        """
        loads and return the WHEEL configuration file
        :return: the loaded configuration file
        """
        return WheelFileConfiguration.load(self.wheel_path())

    def wheel_path(self) -> Path:
        """
        :return: the path to the WHEEL configuration file
        """
        return self.path / "WHEEL"

    def is_app_container(self) -> bool:
        """
        :return: True if this package dist-info is marked as an app-container
        """
        mode_info = self.load_installation_info()
        return mode_info and mode_info.containerized

    def load_entrypoints_cfg(self) -> "EntrypointsConfiguration":
        """
        load and return the entry_points.txt configurate file
        :return: the loaded configuration file
        """
        return EntrypointsConfiguration.load(self.path / "entry_points.txt")

    def load_installation_info(self) -> Optional[PackageInstallationInfo]:
        if self.installation_info_path().exists():
            return PackageInstallationInfo.from_config(json.loads(self.installation_info_path().read_text()))
        return None

    def save_installation_info(self, installation_mode: PackageInstallationInfo):
        self.installation_info_path().write_text(json.dumps(installation_mode.to_config()))

    def load_metadata_cfg(self) -> "PackageMetadata":
        """
        load and return the METADATA configuration file
        :return: the loaded configuration file
        """
        return PackageMetadata.load(self.metadata_path())

    def metadata_path(self) -> Path:
        """
        :return: the path to the METADATA configuration file
        """
        return self.path / "METADATA"

    def load_record_cfg(self) -> "RecordsFileConfiguration":
        """
        load and return the RECORD configuration file
        :return: the loaded configuration file
        """
        return RecordsFileConfiguration.load(self.record_path())

    def record_path(self) -> Path:
        """
        :return: the path to the RECORD configuration file
        """
        return self.path / "RECORD"

    def license_path(self) -> Path:
        """
        :return: the path to the LICENSE file
        """
        return self.path / "LICENSE"

    def installation_info_path(self) -> Path:
        """
        :return: the path to the pkm's added INSTALLATION_MODE file
        """
        return self.path / "INSTALLATION_MODE"

    def user_requested_path(self) -> Path:
        """
        :return: the path to the REQUESTED marker file
        """
        return self.path / "REQUESTED"

    def mark_as_user_requested(self, info: Dependency):
        """
        marks the given package as a user requested one, pkm will not consider it as an orphan package if no other
        package depends on it
        :param info: information about the user request
        """
        self.user_requested_path().write_text(str(info))

    def unmark_as_user_requested(self):
        """
        remove the "user request" mark from a package, pkm will consider it as an orphan package if no other
        package depends on it
        """
        self.user_requested_path().unlink(missing_ok=True)

    def is_user_requested(self) -> bool:
        """
        :return: True if this package dist-info is marked as user-requested
        """
        return self.user_requested_path().exists()

    def load_user_requested_info(self) -> Optional[Dependency]:
        try:
            return Dependency.parse(self.user_requested_path().read_text().strip())
        except:  # noqa
            return None

    @classmethod
    def load(cls, path: Path, non_standard_name_ok: bool = False) -> "DistInfo":
        """
        loads the given `path` as a distinfo
        :param path: the path to load, must be a directory
        :param non_standard_name_ok: if False and the name of the given path is not standard will raise a ValueError
        :return: the loaded distinfo
        """

        if path.exists() and not path.is_dir():
            raise ValueError(f"the given path is not a directory: {path}")

        if path.suffix != '.dist-info' and not non_standard_name_ok:
            raise ValueError(f"{str(path)} is not a properly named dist-info directory")

        return cls(path)

    @classmethod
    def scan(cls, path: Path) -> Iterator[DistInfo]:
        """
        scans a path for all its standard-named dist-info children, load and yield them.
        :param path: the path to scan
        :return: iterator over all the found dist-info children
        """
        for file in path.iterdir():
            if file.suffix == ".dist-info":
                yield DistInfo.load(file)

    @classmethod
    def expected_dir_name(cls, package: "PackageDescriptor") -> str:
        """
        :param package: the package to compute the expected directory name for
        :return: the name of the dist-info directory (last element in the path for the dist-info) that is expected to be
                 created for the given package
        """
        return f"{package.expected_src_package_name}-{str(package.version).replace('-', '_')}.dist-info"

    def installed_files(self) -> Iterator[Path]:
        """
        :return: paths to all the files that were installed to the environment via this package (taken from RECORD)
        """
        root = self.path.parent
        for record in self.load_record_cfg().records:
            file = resolve_relativity(Path(record.file), root).resolve()
            yield file

        yield self.record_path()


@dataclass(frozen=True, eq=True)
class PackageInstallationInfo:
    containerized: bool = False
    editable: bool = False
    compatibility_tag: str = ""

    def to_config(self) -> Dict[str, Any]:
        return {**self.__dict__}

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> PackageInstallationInfo:
        return PackageInstallationInfo(**config)


class EntrypointsConfiguration(IniFileConfiguration):
    entrypoints: List["EntryPoint"]

    @computed_based_on("")
    def entrypoints(self) -> Iterable["EntryPoint"]:

        result: List[EntryPoint] = []

        for group_name, group in self.items():
            for entry_point, object_ref in group.items():
                result.append(EntryPoint(group_name, entry_point, ObjectReference.parse(object_ref)))

        return result

    @entrypoints.modifier
    def set_entrypoints(self, entries: List["EntryPoint"]):
        self._data.clear()
        by_group: Dict[str, List[EntryPoint]] = groupby(entries, key=lambda e: e.group)
        new_data = {group: {e.name: e.ref for e in entries} for group, entries in by_group.items()}
        self._data.update(new_data)


class WheelFileConfiguration(FileConfiguration):

    def generate_content(self) -> str:
        return os.linesep.join(f'{k}: {v}' for k, v in self._data.items())

    def validate_supported_version(self):
        wv = Version.parse(self['Wheel-Version'] or 'unprovided')
        if not isinstance(wv, StandardVersion):
            raise UnsupportedOperationException(f"unknown wheel version: {wv}")

        if wv.release[0] != 1:
            raise UnsupportedOperationException(f"unsupported wheel version: {wv}")
        if wv.release[1] != 0:
            print(f'advanced wheel version: {wv} detected, will be treated as version 1.0')

    @classmethod
    def create(cls, generator: str, purelib: bool):
        return cls(path=None, data={
            'Wheel-Version': '1.0',
            'Generator': generator,
            'Root-Is-Purelib': 'true' if purelib else 'false'
        })

    @classmethod
    def load(cls, path: Path):
        if not path.exists():
            return cls(path=path, data={})

        seperator_rx = re.compile("\\s*:\\s*")
        lines = (line.strip() for line in path.read_text().splitlines())
        kvs = (seperator_rx.split(kv) for kv in lines if kv)
        return cls(path=path, data={kv[0]: kv[1] for kv in kvs})


@dataclass(frozen=True)
class Record:
    file: str
    hash_signature: HashSignature
    length: int

    def absolute_path(self, dist_info: DistInfo) -> Path:
        return resolve_relativity(Path(self.file), dist_info.path).resolve()


class RecordsFileConfiguration(FileConfiguration):

    def generate_content(self) -> str:
        raise UnsupportedOperationException()

    @property
    def records(self) -> List[Record]:
        return self['records']

    def save_to(self, path: Optional[Path] = None):
        path = path or self.path
        path.parent.mkdir(parents=True, exist_ok=True)

        with path.open('w+', newline='') as new_record_fd:
            r: Record
            csv.writer(new_record_fd).writerows(
                (str(r.file), str(r.hash_signature), r.length)
                for r in self['records'])

    def sign_files(self, files: Iterable[Path], root: Path,
                   precomputed_hashes: Optional[Dict[str, HashSignature]] = None) -> RecordsFileConfiguration:
        """
        add to the records in this file the signatures for the given `files`
        :param files: the files to sign
        :param root: a root directory to sign the files relative to, when signing, the record file path will be writen
                     relative to this root
        :param precomputed_hashes: dictionary containing some or all of the given files precomputed hashes,
               its key is the relative path from root to each of the files
        :return: self (for chaining support)
        """

        precomputed_hashes = precomputed_hashes or {}
        records: List[Record] = self.records
        for file in files:
            if file.exists():
                if not file.is_dir():
                    path = str(path_to(root, file))

                    hash_sig = get_or_compute(
                        precomputed_hashes, path,
                        lambda: HashSignature.compute_urlsafe_base64_nopad_encoded('sha256', file))

                    records.append(Record(
                        path,
                        hash_sig,
                        file.lstat().st_size  # size
                    ))
            else:
                warnings.warn(f"requiring to sign non existing file - {file}, skipping it")

        return self

    def sign_recursive(self, content_root: Path) -> RecordsFileConfiguration:
        """
        add to the records in this file the signatures for files inside the `content_root`
        :param content_root: the content to sign
               (will recursively sign all files in the content root and add their signature to the created record file)
        :return: self (for chaining support)
        """
        return self.sign_files(content_root.rglob("*"), content_root)

    @classmethod
    def load(cls, path: Path) -> "RecordsFileConfiguration":
        records = []

        if path.exists():
            with path.open('r', newline='') as records_fd:
                for record in csv.reader(records_fd):
                    file, hash_sig, length = record
                    if hash_sig:
                        records.append(Record(
                            str(Path(file)),  # wrapping in path and then str so that os dependent path will be used
                            HashSignature.parse_urlsafe_base64_nopad_encoded(hash_sig),
                            int(length),
                        ))

        return RecordsFileConfiguration(data={'records': records}, path=path)



import io
from io import IOBase
from typing import IO, Protocol, runtime_checkable, Iterator, Union


# @contextmanager
# def capture(stream: Literal['stdout', 'stderr']) -> ContextManager[StringIO]:
#     original = getattr(sys, stream)
#     try:
#         new_stream = StringIO()
#         setattr(sys, stream, new_stream)
#         yield new_stream
#     finally:
#         setattr(sys, stream, original)

@runtime_checkable
class _HasReadInto(Protocol):
    def readinto(self, b: bytearray) -> int:
        ...


def chunks(io_stream: Union[IO, IOBase], chunk_size: int = io.DEFAULT_BUFFER_SIZE, reuse: bool = True)\
        -> Iterator[bytes]:
    """
    :param io_stream: the stream to iterate on
    :param chunk_size: the required size of each chunk (last chunk may have a different size)
    :param reuse: if True, an attempt to reuse the memory of each chunk is made
                  (read: if reuse is True, don't store the chunks returned by the iterator)
    :return: an iterator on the chunks composing the given `io_stream`
    """
    if reuse and isinstance(io_stream, _HasReadInto):
        chunk = bytearray(chunk_size)
        while sz := io_stream.readinto(chunk):
            if sz == chunk_size:
                yield chunk
            else:
                yield memoryview(chunk)[:sz]
    else:
        while chunk := io_stream.read(chunk_size):
            yield chunk



KNOWN_LICENSES = [
    "Aladdin Free Public License (AFPL)", "CC0 1.0 Universal (CC0 1.0) Public Domain Dedication",
    "DFSG approved", "Eiffel Forum License (EFL)", "Free For Educational Use", "Free For Home Use",
    "Free for non-commercial use", "Freely Distributable", "Free To Use But Restricted", "Freeware",
    "Netscape Public License (NPL)", "Nokia Open Source License (NOKOS)", "Academic Free License (AFL)",
    "Apache Software License", "Apple Public Source License", "Artistic License",
    "Attribution Assurance License", "BSD License", "CEA CNRS Inria Logiciel Libre License, version 2.1 (CeCILL-2.1)",
    "Common Public License", "Eiffel Forum License", "European Union Public Licence 1.0 (EUPL 1.0)",
    "European Union Public Licence 1.1 (EUPL 1.1)", "GNU Affero General Public License v3",
    "GNU Affero General Public License v3 or later (AGPLv3+)", "GNU Free Documentation License (FDL)",
    "GNU General Public License (GPL)", "GNU General Public License v2 (GPLv2)",
    "GNU General Public License v2 or later (GPLv2+)", "GNU General Public License v3 (GPLv3)",
    "GNU General Public License v3 or later (GPLv3+)", "GNU Lesser General Public License v2 (LGPLv2)",
    "GNU Lesser General Public License v2 or later (LGPLv2+)", "GNU Lesser General Public License v3 (LGPLv3)",
    "GNU Lesser General Public License v3 or later (LGPLv3+)", "GNU Library or Lesser General Public License (LGPL)",
    "IBM Public License", "Intel Open Source License", "ISC License (ISCL)", "Jabber Open Source License",
    "MIT License", "MITRE Collaborative Virtual Workspace License (CVW)", "Motosoto License",
    "Mozilla Public License 1.0 (MPL)", "Mozilla Public License 1.1 (MPL 1.1)", "Mozilla Public License 2.0 (MPL 2.0)",
    "Nethack General Public License", "Nokia Open Source License", "Open Group Test Suite License",
    "Python License (CNRI Python License)", "Python Software Foundation License", "Qt Public License (QPL)",
    "Ricoh Source Code Public License", "Sleepycat License", "Sun Industry Standards Source License (SISSL)",
    "Sun Public License", "University of Illinois/NCSA Open Source License", "Vovida Software License 1.0",
    "W3C License", "X.Net License", "zlib/libpng License", "Zope Public License", "Other/Proprietary License",
    "Public Domain", "Repoze Public License", ]



import os
import platform
import subprocess
import sys
from dataclasses import dataclass
from subprocess import CompletedProcess
from typing import List, NoReturn, Mapping

from pkm.utils.monitors import MonitoredOperation, MonitoredEvent
from pkm.utils.parsers import SimpleParser

_IS_WINDOWS = platform.system() == 'Windows'


def execvpe(cmd: str, args: List[str], env: Mapping[str, str]) -> NoReturn:
    if _IS_WINDOWS:
        sys.exit(subprocess.run([cmd, *args], env=env).returncode)

    os.execvpe(cmd, [cmd, *args], env)


def monitored_run(execution_name: str, cmd: List[str], **subprocess_run_kwargs) -> CompletedProcess:
    with ProcessExecutionMonitoredOp(execution_name, cmd) as mpo:
        subprocess_run_kwargs['stdout'] = subprocess.PIPE
        subprocess_run_kwargs['stderr'] = subprocess.STDOUT
        proc = subprocess.Popen(cmd, **subprocess_run_kwargs)
        stdout = proc.stdout

        def decode(line):  # noqa
            nonlocal decode
            if hasattr(line, 'decode'):
                decode = lambda line: line.decode()  # noqa
            else:
                decode = lambda line: line  # noqa

            return decode(line)

        while (line := stdout.readline()) or (rcode := proc.poll()) is None:
            ProcessExecutionOutputLineEvent(decode(line).strip()).notify(mpo)

        ProcessExecutionExitEvent(rcode).notify(mpo)
        return CompletedProcess(cmd, rcode)


class _CommandParser(SimpleParser):
    def read_command(self):
        result: List[str] = []

        self.match_ws()
        while self.is_not_empty():
            if self.peek() in "\"'":
                result.append(self.read_str())
            else:
                result.append(self.until(lambda i, s: s[i].isspace()))
            self.match_ws()

        return result

    def read_str(self) -> str:
        quotes = self.match_any('"', "'")
        if not quotes:
            self.raise_err("expecting start of string")

        s = quotes + self.until_match(quotes)
        while s[-1] == '\\':
            s += self.next() + self.until_match(quotes)

        s += quotes
        self.match_or_err(quotes, 'basic string ending expected')

        return eval(s)


def split_command(cmd: str) -> List[str]:
    return _CommandParser(cmd).read_command()


@dataclass
class ProcessExecutionMonitoredOp(MonitoredOperation):
    execution_name: str
    cmd: List[str]


@dataclass
class ProcessExecutionOutputLineEvent(MonitoredEvent):
    line: str


@dataclass
class ProcessExecutionExitEvent(MonitoredEvent):
    exit_code: int



from __future__ import annotations

import dataclasses
import hashlib
import os
import sys
from collections import defaultdict
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from subprocess import CompletedProcess
from typing import List, Set, Dict, Optional, Union, TypeVar, NoReturn, MutableMapping, TYPE_CHECKING

from pkm.api.dependencies.dependency import Dependency
from pkm.api.distributions.pth_link import PthLink
from pkm.api.environments.environment_introspection import EnvironmentIntrospection
from pkm.api.packages.package_installation import PackageInstallationTarget
from pkm.api.packages.site_packages import SitePackages
from pkm.api.repositories.repository import Repository, HasAttachedRepository
from pkm.api.versions.version import StandardVersion, Version
from pkm.utils.commons import unone, NoSuchElementException
from pkm.utils.entrypoints import EntryPoint
from pkm.utils.files import is_root_path
from pkm.utils.iterators import find_first
from pkm.utils.processes import execvpe, monitored_run
from pkm.utils.properties import cached_property, clear_cached_properties
from pkm.utils.types import SupportsLessThanEq

if TYPE_CHECKING:
    from pkm.api.environments.environments_zoo import EnvironmentsZoo
    from pkm.api.environments.containerized_apps import ContainerizedApplications
    from pkm.api.repositories.repository_management import RepositoryManagement

_DEPENDENCIES_T = Union[Dependency, str, List[Union[Dependency, str]]]
_PACKAGE_NAMES_T = Union[str, List[str]]
_T = TypeVar("_T")


class Environment(HasAttachedRepository):

    def __init__(self, env_path: Path, interpreter_path: Optional[Path] = None, *,
                 use_user_site: bool = False, zoo: Optional["EnvironmentsZoo"] = None):
        self._env_path = env_path
        self._interpreter_path = interpreter_path
        self._use_user_site = use_user_site

        if zoo:
            self.zoo = zoo  # noqa

    @property
    def path(self) -> Path:
        """
        :return: the path for this environment root directory
        """
        return self._env_path

    @cached_property
    def zoo(self) -> Optional["EnvironmentsZoo"]:
        from pkm.api.environments.environments_zoo import EnvironmentsZoo
        if EnvironmentsZoo.is_valid(zoo_path := self.path.parent):
            return EnvironmentsZoo.load(zoo_path)
        return None

    @property
    def app_containers(self) -> "ContainerizedApplications":
        return self.installation_target.app_containers

    @cached_property
    def repository_management(self) -> "RepositoryManagement":
        from pkm.api.repositories.repository_management import EnvRepositoryManagement
        return EnvRepositoryManagement(self)

    @cached_property
    def _introspection(self) -> EnvironmentIntrospection:
        if is_root_path(self._env_path):
            return EnvironmentIntrospection.compute(self.interpreter_path)
        return EnvironmentIntrospection.load_or_compute(
            self._env_path / 'etc/pkm/env_introspection.json', self.interpreter_path, True)

    @property
    def name(self) -> str:
        return self.path.name

    @cached_property
    def installation_target(self) -> PackageInstallationTarget:
        paths = self._introspection.user_paths() if self._use_user_site else self._introspection.paths
        if not paths:
            raise NoSuchElementException("could not find user site configuration matching to the current os")

        result = PackageInstallationTarget(self, **{
            field.name: paths.get(field.name) or ""
            for field in dataclasses.fields(PackageInstallationTarget)
            if field.name != 'env'
        })

        return result

    def __repr__(self):
        return f"Environment({self.path})"

    @cached_property
    def interpreter_version(self) -> StandardVersion:
        """
        :return: the version of the environment's python interpreter
        """
        return StandardVersion(release=tuple(self._introspection.interpreter_version)[:3])

    @cached_property
    def interpreter_path(self) -> Path:
        """
        :return: the path for the environment's python interpreter
        """
        if self._interpreter_path is None:
            self._interpreter_path = _find_interpreter(self._env_path)
            if not self._interpreter_path:
                raise ValueError(f"could not determine the environment interpreter path for env: {self.path}")
        return self._interpreter_path

    @cached_property
    def operating_platform(self) -> "OperatingPlatform":
        emarkers = self.markers
        ispc = self._introspection

        return OperatingPlatform(
            os=emarkers["platform_system"].lower(),
            os_bits=ispc['os.bits'],
            machine=emarkers["platform_machine"].lower(),
            python_version=self.interpreter_version,
            python_implementation=emarkers["platform_python_implementation"].lower(),
            python_interpreter_bits=32 if ispc.is_32bit_interpreter else 64
        )

    def compatibility_tag_score(self, tag: str) -> Optional[SupportsLessThanEq]:
        """
        compute the compatibility score for the given pep425 compatibility tag
        :param tag: the pep425 compatibility tag
        :return: an opaque score object that support __le__ and __eq__ operations (read: comparable)
                 which can be treated as a score (read: higher is better)
        """
        return self._introspection.compatibility_score(tag)

    @cached_property
    def markers(self) -> Dict[str, str]:
        """
        :return: pep508 environment markers  
        """
        return self._introspection.compute_markers()

    @cached_property
    def site_packages(self) -> SitePackages:
        return self.installation_target.site_packages

    @cached_property
    def markers_hash(self) -> str:
        """
        :return: a hash built from the environment's markers, can be used to identify instances of this environment
        """
        sorted_markers = sorted(self.markers.items(), key=lambda item: item[0])
        marker_str = ';'.join(f"{k}={v}" for k, v in sorted_markers)
        return hashlib.md5(marker_str.encode()).hexdigest()

    @contextmanager
    def activate(self, env: MutableMapping[str, str] = os.environ):
        new_path = f"{self._introspection.paths['scripts']}"
        if old_path := env.get("PATH"):
            new_path = f"{new_path}{os.pathsep}{old_path}"

        old_venv = env.get('VIRTUAL_ENV')
        new_venv = str(self.path)

        env["PATH"] = new_path
        env["VIRTUAL_ENV"] = new_venv

        try:
            yield
        finally:
            if old_path is not None:
                env["PATH"] = old_path
            else:
                del env["PATH"]

            if old_venv is not None:
                env["VIRTUAL_ENV"] = old_venv
            else:
                del env["VIRTUAL_ENV"]

    def exec_proc(self, cmd: str, args: Optional[List[str]] = None, env: Optional[Dict[str, str]] = None) -> NoReturn:
        """
        similar to `run_proc` but does not return, this process will become the new process
        (on supporting operating systems)

        :param cmd: the command to execute
        :param args: list of arguments
        :param env: environment variables
        """

        with self.activate(env):
            execvpe(cmd, args, env)

    def run_proc(self, args: List[str], **subprocess_run_kwargs) -> CompletedProcess:
        """
        execute the given command in a new process, the process will be executed with its path adjusted to include this
        venv binaries and scripts

        :param args: the command to execute
        :param subprocess_run_kwargs: any argument that is accepted by `subprocess.run` (aside from args)
        :return: a `CompletedProcess` instance describing the completion of the requested process
        """
        env = {}
        if extra_env := subprocess_run_kwargs.get('env'):
            env.update(extra_env)
        else:
            env.update(os.environ)

        subprocess_run_kwargs['env'] = env
        with self.activate(env):
            return monitored_run(args[0], args, **subprocess_run_kwargs)

    def reload(self):
        """
        reload volatile information about this environment (like the installed packages)
        """
        clear_cached_properties(self)

    def install_link(self, name: str, paths: List[Path], imports: Optional[List[str]] = None):
        """
        installs a pth link (named `name`.pth) in the site packages (purelib)
        :param name:
        :param imports: imports to execute at runtime when reading this link
                        (example: `imports` = ['abc'] will import the abc package when the interpreter attached to this
                        environment gets executed and meets the created link)
        :param paths: the paths to add in the created link
        """

        imports = unone(imports, list)
        pth_file = self.site_packages.purelib_path / f"{name}.pth"
        if pth_file.exists():
            raise FileExistsError(f"the file {pth_file} already exists")

        PthLink(pth_file, paths, imports).save()

    def install(
            self, dependencies: _DEPENDENCIES_T, repository: Optional[Repository] = None, user_requested: bool = True,
            dependencies_override: Optional[Dict[str, List[Dependency]]] = None,
            editables: Optional[Dict[str, bool]] = None, updates: Optional[List[str]] = None):
        """
        installs the given set of dependencies into this environment.
        see: `prepare_installation` for more information about this method arguments
        """

        self.installation_target.install(
            _coerce_dependencies(dependencies), repository, user_requested, dependencies_override, editables,
            updates)

    def force_remove(self, package: str):
        """
        forcefully remove the required package, will not remove its dependencies and will not check if other packages
        depends on it - use this method with care (or don't use it at all :) )
        :param package: the name of the package to be removed
        """
        self.installation_target.force_remove(package)

    def uninstall(self, packages: _PACKAGE_NAMES_T) -> Set[str]:
        """
        attempt to remove the required packages from this env together will all the dependencies that may become orphan
        as a result of this step.

        if a package `p in packages` is a dependency (directly or indirectly) of another
        "user requested" package `q not in packages` then `p` will be kept in the environment but its
        "user requested" flag will be removed (if it was existed)

        :param packages: the package names to remove
        :return the set of package names that were successfully removed from the environment
        """
        return self.installation_target.uninstall(_coerce_package_names(packages))

    @cached_property
    def entrypoints(self) -> Dict[str, List[EntryPoint]]:
        """
        :return: all entrypoints in this environment grouped by their defined group
        """
        groups: Dict[str, List[EntryPoint]] = defaultdict(list)

        for package in self.site_packages.installed_packages():
            for ep in package.dist_info.load_entrypoints_cfg().entrypoints:
                groups[ep.group].append(ep)

        return groups

    @staticmethod
    def is_valid(path: Path) -> bool:
        """
        :param path: a path that may contain a python environment
        :return: true if this path contains a python environment
        """
        return (path / "pyvenv.cfg").exists() and _find_interpreter(path) is not None

    @classmethod
    def of_interpreter(cls, interpreter: Path, site: str = "user") -> Environment:
        """
        load an environment using the given interpreter
        :param interpreter: the interpreter to use
        :param site: control the installation site of the returned environment - acceptable values are
            'user' and 'system', it is only applicable for system (read: non-virtual) environments. the 'system' option
            just use the regular site, you can read about the 'user' site in documentation of
            `site.getusersitepackages()`
        :return: the loaded environment
        """
        if (interpreter.parent.parent / "pyvenv.cfg").exists():
            return cls(interpreter.parent.parent, interpreter)
        else:  # this is a system environment
            return cls(Path("/"), interpreter, use_user_site=site == 'user')

    @classmethod
    def current(cls, site: str = "user") -> Environment:
        """
        load the environment used by the currently executed interpreter
        :param site: control the installation site of the returned environment - acceptable values are
            'user' and 'system', it is only applicable for system (read: non-virtual) environments. the 'system' option
            just use the regular site, you can read about the 'user' site in documentation of
            `site.getusersitepackages()`
        :return: the loaded environment
        """
        return cls.of_interpreter(Path(sys.executable), site)

    @classmethod
    def load(cls, path: Union[Path, str]) -> Environment:
        return cls(Path(path))


def _coerce_dependencies(dependencies: _DEPENDENCIES_T) -> List[Dependency]:
    if isinstance(dependencies, str):
        return [Dependency.parse(dependencies)]
    if isinstance(dependencies, Dependency):
        return [dependencies]
    return [cd for dep in dependencies for cd in _coerce_dependencies(dep)]


def _coerce_package_names(package_names: _PACKAGE_NAMES_T) -> List[str]:
    if isinstance(package_names, str):
        return [package_names]
    return package_names


def _find_interpreter(env_root: Path) -> Optional[Path]:
    return find_first((env_root / "bin/python", env_root / "Scripts/python.exe"), lambda it: it.exists())


@dataclass
class OperatingPlatform:
    os: str  # 'linux', 'darwin', 'java', 'windows' - the value returned by platform.system() in lowercase
    os_bits: int  # 32 or 64
    machine: str  # the value returned by platform.machine() in lowercase
    python_version: Version
    python_implementation: str  # the value returned by platform.python_implementation() in lowercase
    python_interpreter_bits: int  # 32 or 64

    def has_windows_os(self) -> bool:
        return self.os == 'Windows'

    def has_x86_cpu(self) -> bool:
        return self.machine in ("i386", "amd64", "x86", "x86_64")

    def has_arm_cpu(self) -> bool:
        return 'armv71' in self.machine or 'aarch64' in self.machine



from abc import ABC, abstractmethod
from concurrent.futures import Future, ThreadPoolExecutor
from threading import Lock, Condition
from typing import TypeVar, Generic, Callable, Any, Optional, List
import typing

from pkm.utils.types import ParamSpec

_T = TypeVar("_T")
_U = TypeVar("_U")
_PARAMS: "typing.ParamSpec" = ParamSpec("_PARAMS")


class Promise(ABC, Generic[_T]):

    @abstractmethod
    def when_completed(self, callback: "Callable[[Promise[_T]], _U]") -> "Promise[_U]":
        ...

    @abstractmethod
    def result(self, wait: bool = True) -> _T:
        ...

    @abstractmethod
    def is_completed(self) -> bool:
        ...

    @abstractmethod
    def is_failed(self) -> bool:
        ...

    @abstractmethod
    def request_cancel(self):
        ...

    def is_succeeded(self) -> bool:
        return self.is_completed() and not self.is_failed()

    @classmethod
    def wrap_future(cls, future: Future, result_mapper: Callable[[Future], _T]) -> "Promise[_T]":
        return _FutureWrappingPromise(future, result_mapper)

    @classmethod
    def execute(cls, executor: ThreadPoolExecutor, mtd: Callable[_PARAMS, _T],
                *args: _PARAMS.args, **kwargs: _PARAMS.kwargs):
        return cls.wrap_future(executor.submit(mtd, *args, **kwargs), lambda fu: fu.result())

    @classmethod
    def create_completed(cls, result: Optional[_T] = None, err: Optional[Exception] = None):
        promise = _SimplePromise()

        if err is None:
            # noinspection PyProtectedMember
            promise._complete(result)
        else:
            # noinspection PyProtectedMember
            promise._fail(err)

        return promise


class Deferred(Generic[_T]):

    def __init__(self):
        outer = self

        class InnerSimplePromise(_SimplePromise[_T]):
            def request_cancel(self):
                outer._cancel_requested = True

        self._promise: InnerSimplePromise[_T] = InnerSimplePromise()
        self._cancel_requested: bool = False

    def promise(self) -> Promise[_T]:
        return self._promise

    def is_cancel_requested(self) -> bool:
        return self._cancel_requested

    def complete(self, result: _T):
        # noinspection PyProtectedMember
        self._promise._complete(result)

    def fail(self, failure: Exception):
        # noinspection PyProtectedMember
        self._promise._fail(failure)


def _default_future_wrapping_promise_result_mapper(fu: Future) -> Any:
    return fu.result()


class _SimplePromise(Promise[_T]):

    def __init__(self):
        self._wait_lock = Condition(Lock())
        self._result: Optional[_T] = None
        self._err: Optional[Exception] = None
        self._done: bool = False
        self._listeners: List[Callable[[], None]] = []

    def when_completed(self, callback: "Callable[[Promise[_T]], _U]") -> "Promise[_U]":

        with self._wait_lock:
            if not self._done:

                deferred = Deferred()

                def callback_wrapper():
                    # noinspection PyShadowingNames
                    try:
                        deferred.complete(callback(self))
                    except Exception as e:
                        deferred.fail(e)

                self._listeners.append(callback_wrapper)
                return deferred.promise()

        try:
            return Promise.create_completed(result=callback(self))
        except Exception as e:
            return Promise.create_completed(err=e)

    def result(self, wait: bool = True) -> _T:
        while True:
            with self._wait_lock:
                if self._done:
                    if self._err:
                        raise self._err
                    return self._result

                if not wait:
                    raise ValueError("promise not completed yet")

                self._wait_lock.wait()

    def is_completed(self) -> bool:
        return self._done

    def is_failed(self) -> bool:
        return self._done and self._err

    def request_cancel(self):
        pass  # unsupported..

    def _complete(self, result: _T):
        with self._wait_lock:
            if self._done:
                raise ValueError("already completed")
            self._result = result
            self._done = True
            self._wait_lock.notify_all()

    def _fail(self, err: Exception):
        if err is None:
            raise ValueError("exception must be given")

        with self._wait_lock:
            if self._done:
                raise ValueError("already completed")

            self._err = err
            self._done = True
            self._wait_lock.notify_all()


class _FutureWrappingPromise(_SimplePromise[_T]):
    def __init__(self, future: Future,
                 result_mapper: Callable[[Future], _T] = _default_future_wrapping_promise_result_mapper):

        super().__init__()
        self._future = future

        def done_callback(fu: Future):
            try:
                self._complete(result_mapper(fu))
            except Exception as e:
                self._fail(e)

        future.add_done_callback(done_callback)

    def request_cancel(self):
        self._future.cancel()



import inspect
from contextlib import contextmanager
from dataclasses import dataclass
from typing import Any, Callable, Type, Dict, Optional, Generator

from pkm.utils.dicts import get_or_put
from pkm.utils.properties import cached_property
from pkm.utils.strings import startswith_any


@dataclass
class _Callback:
    parent: Optional["_Callback"] = None
    enter_method: Optional[Callable] = None
    leave_method: Optional[Callable] = None
    fail_method: Optional[Callable] = None


# noinspection PyUnusedLocal
class _Monitor:

    @cached_property
    def _callbacks(self) -> Dict[Type, _Callback]:
        return {}

    def add_listeners(self, **callbacks):
        new_callbacks: Dict[Type, _Callback] = self._create_callbacks(callbacks)
        for type_, callback in new_callbacks.items():
            self._callbacks[type_] = callback

    @contextmanager
    def listen(self, **callbacks):

        method_name: str
        new_callbacks: Dict[Type, _Callback] = self._create_callbacks(callbacks)
        for type_, callback in new_callbacks.items():
            self._callbacks[type_] = callback

        try:
            yield
        finally:
            for type_, callback in new_callbacks.items():
                self._callbacks[type_] = callback.parent

    def _create_callbacks(self, callbacks: Dict[str, Any]) -> Dict[Type, _Callback]:
        new_callbacks: Dict[Type, _Callback] = {}
        for method_name, method in callbacks.items():
            if (mode := startswith_any(method_name, ("enter_", "leave_", "on_", "failed_", "with_"))) \
                    and isinstance(method, Callable):

                arg_inspection = inspect.getfullargspec(method)
                if len(arg_inspection.args) == 0:
                    continue

                first_arg = arg_inspection.args[0]
                if first_arg_type := arg_inspection.annotations.get(first_arg):
                    callback: _Callback = get_or_put(
                        new_callbacks, first_arg_type,
                        lambda: _Callback(self._callbacks.get(first_arg_type)))

                    if mode in ("enter_", "on_"):
                        callback.enter_method = method
                    elif mode == "leave_":
                        callback.leave_method = method
                    elif mode == "failed_":
                        callback.fail_method = method
                    else:  # mode == with_

                        def create(method_):
                            coroutines: Dict[int, Generator] = {}

                            def enter(x, *args, **kwargs):
                                if isinstance(coroutine := method_(x, *args, **kwargs), Generator):
                                    try:
                                        coroutine.__next__()
                                        coroutines[id(x)] = coroutine
                                    except StopIteration:
                                        pass

                            def leave(x, *args, **kwargs):
                                if coroutine := coroutines.pop(id(x), None):
                                    try:
                                        coroutine.__next__()
                                    except StopIteration:
                                        pass

                            def fail(x: Any, err: Exception, *args, **kwargs):
                                if coroutine := coroutines.pop(id(x), None):
                                    coroutine.throw(err)

                            callback.enter_method = enter
                            callback.leave_method = leave
                            callback.fail_method = fail

                        create(method)
        return new_callbacks

    @contextmanager
    def enter(self, context: Any):

        cb = self._callbacks.get(type(context))
        if isinstance(context, _Monitor):
            context._parent_monitor = self

        try:
            cb and cb.enter_method and cb.enter_method(context)
            yield
        except Exception as e:
            cb and cb.fail_method and cb.fail_method(context, e)
        finally:
            cb and cb.leave_method and cb.leave_method(context)

    def notify(self, event: Any):
        (cb := self._callbacks.get(type(event))) and cb.enter_method(event)


Monitor = _Monitor()


class MonitoredEvent:
    def notify(self, monitor: _Monitor = Monitor):
        monitor.notify(self)


class MonitoredOperation(_Monitor):
    def __enter__(self):
        # noinspection PyProtectedMember
        if hasattr(self, '_parent_monitor'):
            parent = self._parent_monitor
        else:
            parent = Monitor

        cb = self._callback = parent._callbacks.get(type(self))

        cb and cb.enter_method and cb.enter_method(self)

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        cb = self._callback

        if exc_type:
            cb and cb.fail_method and cb.fail_method(self, exc_val)

        cb and cb.leave_method and cb.leave_method(self)



from dataclasses import dataclass
from pathlib import Path
from typing import Optional, Dict, Any

from pkm.config.configuration import TomlFileConfiguration, computed_based_on

ENVIRONMENT_CONFIGURATION_PATH = "etc/pkm/environments.toml"


@dataclass(frozen=True, eq=True)
class AttachedEnvironmentConfig:
    path: Optional[Path] = None
    zoo: Optional[Path] = None

    def to_config(self) -> Dict[str, Any]:
        return {k: str(v) for k, v in self.__dict__ if v is not None}

    @classmethod
    def from_config(cls, cfg: Dict[str, Any]):
        return AttachedEnvironmentConfig(**{k: Path(v) for k, v in cfg.items()})


class EnvironmentsConfiguration(TomlFileConfiguration):
    attached_env: Optional[AttachedEnvironmentConfig]

    @computed_based_on("attached-env")
    def attached_env(self) -> Optional[AttachedEnvironmentConfig]:
        if config := self["attached-env"]:
            return AttachedEnvironmentConfig.from_config(config)
        return None



import os
from typing import Optional, NoReturn, Callable, List, Type, TypeVar

_URL_CHARS = set("ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-._~:/?#[]@!$&'()*+,;=%")
_P = TypeVar("_P", bound="SimpleParser")
_T = TypeVar("_T")


class SimpleParser:
    """
    simple base class for parsers.
    functions semantics:
        match* -> return the matched data, if nothing matched return the empty string
        read* -> return a complex object resulted by parsing the content,
                 if cannot create the requested object raise error
    """

    def __init__(self, text: str, file_name: Optional[str] = None):
        self.text = text
        self.file_name = file_name
        self.position = 0

    def peek(self, amount: int = 1) -> str:
        p = self.position
        return self.text[p:p + amount]

    def is_not_empty(self) -> bool:
        return self.position < len(self.text)

    def is_empty(self) -> bool:
        return self.position == len(self.text)

    def read_digits(self) -> str:
        position = self.position
        self.until(lambda i, t: not t[i].isdigit())
        if self.position == position:
            self.raise_err('expecting digits')
        return self.text[position:self.position]

    def subparser(self, pareser_t: Type[_P]) -> _P:
        main_parser = self
        p = self.position

        class SubParser(pareser_t):
            @property
            def position(self) -> int:
                return main_parser.position

            @position.setter
            def position(self, v: int):
                main_parser.position = v

        result = SubParser(text=self.text)
        self.position = p
        return result

    def read_url(self, supported_schemas: Optional[List[str]] = None) -> str:
        p = self.position

        supported_schemas = supported_schemas or ['http', 'https']
        if not self.match_any(*supported_schemas):
            self.raise_err('expecting url')

        self.match_or_err('://', 'malformed url')
        self.until(lambda i, t: t[i] not in _URL_CHARS)

        return self.text[p:self.position]

    def match(self, *substrs: str) -> Optional[str]:
        p = self.position
        for i, substr in enumerate(substrs):
            if i != 0:
                self.match_ws(False)

            if self.peek(len(substr)) != substr:
                self.position = p
                return None

            self.position += len(substr)

        return self.text[p:self.position]

    def peek_prev(self, amount: int = 1) -> str:
        p = self.position
        start = max(0, p - amount)
        return self.text[start:p]

    def match_any(self, *substrs: str) -> Optional[str]:
        for substr in substrs:
            if self.match(substr):
                return substr

        return None

    def raise_err(self, msg: str, lines_to_show: int = 4) -> NoReturn:
        lines = self.text.splitlines(keepends=True)
        current_line = self.text.count('\n', 0, self.position)
        start_line = max(0, current_line - lines_to_show)
        sub_line_len = self.position - sum(len(it) for it in lines[:current_line])
        largest_line_len = max(len(lines[i]) if i < len(lines) else 0 for i in range(start_line, current_line + 1))
        pref = f'... after {start_line} lines ...\n\n' if start_line > 0 else ''
        largest_line_len = max(largest_line_len, len(pref))

        position_indicator = f"AT LINE {current_line}"
        if self.file_name:
            position_indicator = f"{self.file_name}:{current_line}"

        lines = ''.join(lines[start_line:current_line + 1])
        if not lines.endswith('\n'):
            lines += '\n'

        out = f"Parsing Failed: {msg} ({position_indicator})\n" \
              f"{'-' * largest_line_len}\n" \
              f"{pref}" \
              f"{lines}" \
              f"{'~' * sub_line_len}^"

        raise ValueError(out)

    def until_match(self, substr: str) -> str:
        pos = self.position

        try:
            self.position = self.text.index(substr, pos)
            return self.text[pos: self.position]
        except ValueError:
            return ''

    def until(self, predicate: Callable[[int, str], bool]) -> str:
        """
        match chars from this parser until `predicate` return True
        :param predicate: the stop condition to evaluate
        :return: the matched chars
        """
        p = self.position
        for i in range(self.position, len(self.text)):
            if predicate(i, self.text):
                self.position = i
                return self.text[p:i]

        self.position = len(self.text)
        return self.text[p:self.position]

    def next(self, amount: int = 1) -> str:
        peek = self.peek(amount)
        self.position += len(peek)

        return peek

    def match_or_err(self, substr: str, err: str) -> None:
        if not self.match(substr):
            self.raise_err(err)

    def __str__(self):
        return f"TextReader(pos={self.position}, '{self.text[self.position: self.position + 25]}...')"

    def __repr__(self):
        return str(self)

    def match_line(self, include_line_end: bool = True) -> str:
        result = self.until_match(os.linesep)
        if include_line_end:
            result += os.linesep
            self.position += len(os.linesep)
        return result

    def read_blank_line(self, include_line_end: bool = True):
        if content := self.match_line(include_line_end).strip():
            self.raise_err(f'unexpected content: {content}')

    def match_ws(self, allow_new_lines: bool = True) -> str:
        p = self.position
        while self.is_not_empty():
            n = self.peek()
            if n == '\n' and not allow_new_lines:
                break
            elif not n.isspace():
                break

            self.next()

        return self.text[p:self.position]



from pathlib import Path
from typing import List, Optional, Union


class EtcChain:
    """
    represents a recursive etc chain finder.
    the idea is that for some applications we want to have "contextual enrichment" for configuration
    e.g., say that we have some configurations file called: x stored in /etc/app-identifier/x
    and in that file there is a project relevant setting p, we want to enable the project to overwrite/merge/remove p,
    so we can add, under the project dir, the file: etc/app-identifier/x with our modifications.
    Now, if, the configuration loader, when reading x, use the etc chain from the root of the project, it will get
    all the "x" files that should affect the configuration.

    It works by finding all instances of the file inside the given context of its parent directories,
    and finally also include the main (/etc/app-identifier/x) configuration in case it wasn't already
    """

    def __init__(self, main: Path, identifier: str):
        self._main = main
        self._identifier = identifier

    def main_config(self, etc_subpath: Union[Path, str]) -> Optional[Path]:
        return self._main / etc_subpath

    def config_chain(
            self, context_path: Path, etc_subpath: Union[str, Path] = "", include_main: bool = True) -> List[Path]:
        # TODO: document order: context to root
        result = []

        search_list = [context_path, *context_path.parents]
        for s in search_list:
            if (sp := s / 'etc' / self._identifier / etc_subpath).exists():
                result.append(sp)

        if (main_subpath := self._main / etc_subpath).exists():
            if include_main and main_subpath not in result:
                result.append(main_subpath)
            elif not include_main and main_subpath in result:
                result.remove(main_subpath)

        return result



import os
import stat
import sys
from pathlib import Path
from typing import TYPE_CHECKING

from pkm.api.distributions.distinfo import EntryPoint
from pkm.launchers import build_windows_script_launcher

if TYPE_CHECKING:
    from pkm.api.environments.environment import Environment


class Executables:

    @staticmethod
    def patch_shabang_for_interpreter(source: Path, target: Path, interpreter_path: Path):
        """
        copy the `source` script to the `target` path, patching the shabng line so that it will be executable
        under the given `env`
        :param source: path to the source script
        :param target: path to put the patched script
        :param interpreter_path: the interpreter which the patched script should be executeable with
        """
        with source.open('rb') as script_fd, target.open('wb+') as target_fd:
            first_line = script_fd.readline()
            if first_line.startswith(b"#!python"):
                w = 'w' if first_line.startswith(b"#!pythonw") else ''
                target_fd.write(
                    f"#!{interpreter_path.absolute()}{w}{os.linesep}".encode(sys.getfilesystemencoding()))
            else:
                target_fd.write(first_line)

            target_fd.write(script_fd.read())

        st = os.stat(source)
        os.chmod(target, st.st_mode | stat.S_IEXEC)

    @staticmethod
    def generate(env: "Environment", target_dir: Path, file_name: str, script: str, is_gui: bool = False):
        """
        generates executable for the given `script` that is runnable under the given environment
        :param env: the environment to be used by the generated executable
        :param target_dir: where to put the executable
        :param file_name: the base name of the script to create (extension may vary)
        :param script: the script content to execute
        :param is_gui: if True, and in windows platform, will create a "gui" script, defaults to False.
        :return: path to the created executable
        """

        if env.operating_platform.has_windows_os():
            return build_windows_script_launcher(env, target_dir, file_name, script, is_gui)
        else:
            source = f"#!{env.interpreter_path.absolute()}\n{script}"

            result_script_path = target_dir / file_name
            result_script_path.write_text(source)

            st = os.stat(result_script_path)
            os.chmod(result_script_path, st.st_mode | stat.S_IEXEC)
            return result_script_path

    @staticmethod
    def generate_for_entrypoint(env: "Environment", entrypoint: EntryPoint, target_dir: Path) -> Path:
        """
        generates executable for the given entrypoint that is runnable under the given environment
        :param entrypoint: the entrypoint to generate the executable for
        :param env: the environment to be used by the generated executable
        :param target_dir: where to put the executable
        :return: path to the created executable
        """

        return Executables.generate(
            env, target_dir, entrypoint.name, entrypoint.ref.execution_script_snippet(),
            entrypoint.group == EntryPoint.G_CONSOLE_SCRIPTS)



from __future__ import annotations

from pathlib import Path
from typing import Iterable, Union, Optional, Tuple, List, TYPE_CHECKING

from pkm.api.projects.environments_config import EnvironmentsConfiguration, ENVIRONMENT_CONFIGURATION_PATH
from pkm.api.projects.project import Project
from pkm.api.repositories.repository import Repository, Authentication, HasAttachedRepository
from pkm.config.configuration import TomlFileConfiguration, computed_based_on
from pkm.utils.files import path_to, ensure_exists, resolve_relativity
from pkm.utils.iterators import single_or_raise
from pkm.utils.properties import cached_property

if TYPE_CHECKING:
    from pkm.api.repositories.repository_management import RepositoryManagement


class ProjectGroup(HasAttachedRepository):
    """
    project group, like the name implies, is a group of projects
    the projects are related to themselves in a parent/children manner,
    a parent is always a group and children may be projects or another project groups

    the project group is configured by the pyproject-group.toml configuration file
    """

    def __init__(self, config: "PyProjectGroupConfiguration", parent: Optional[ProjectGroup] = None):
        self._config = config
        self.path = self._config.path.parent
        if parent:
            self.parent = parent  # noqa

    @property
    def config(self) -> "PyProjectGroupConfiguration":
        return self._config

    @cached_property
    def environments_config(self) -> EnvironmentsConfiguration:
        return EnvironmentsConfiguration.load(self.path / ENVIRONMENT_CONFIGURATION_PATH)

    @cached_property
    def repository_management(self) -> "RepositoryManagement":
        from pkm.api.repositories.repository_management import ProjectGroupRepositoryManagement
        return ProjectGroupRepositoryManagement(self)

    @cached_property
    def parent(self) -> Optional["ProjectGroup"]:
        """
        :return: the parent of this project group (if such exists)
        """
        if not (parent := self._config.parent):
            return ProjectGroup._find_parent(self.path.resolve())
        else:
            return ProjectGroup(PyProjectGroupConfiguration.load(
                ensure_exists(parent, lambda: f"{self.path}'s parent path {parent} doesn't exists")
            ))

    @cached_property
    def root(self) -> Optional["ProjectGroup"]:
        """
        :return: the top most parent of this project group
        """
        return parent.root if (parent := self.parent) else None

    @cached_property
    def children(self) -> Iterable[Union[Project, "ProjectGroup"]]:
        """
        :return: the child projects (or project groups) defined in this group
        """
        result = []
        for child in self._config.children:
            if (child / 'pyproject.toml').exists():
                result.append(Project.load(child, group=self))
            elif (child / 'pyproject-group.toml').exists:
                result.append(
                    ProjectGroup(PyProjectGroupConfiguration.load(child / 'pyproject-group.toml'), parent=self))

        return result

    @cached_property
    def project_children_recursive(self) -> List[Project]:
        """
        :return: the child projects defined in this group and
        recursively the child projects of the project groups defined in this group
        """
        result = []
        for child in self.children:
            if isinstance(child, Project):
                result.append(child)
            else:
                result.extend(child.project_children_recursive)

        return result

    def add(self, project: Union[Project, Path]):
        """
        add project to this group, saving the modification in the configuration file
        :param project: the project to add
        """

        project_path = project if isinstance(project, Path) else project.path
        self._config.children = (*self._config.children, project_path)
        self._config.save()

    def remove(self, project: Union[str, Path]):
        """
        remove project from this group, saving the modification in the configuration file
        :param project: the project name or path to remove
        """
        if isinstance(project, str):
            project = single_or_raise(p for p in self.children if p.name == project).path

        project = project.resolve()

        self._config.children = tuple(p for p in self._config.children if p != project)
        self._config.save()

    def build_all(self):
        """
        recursively run the build operation on all projects and subprojects in this group
        """
        # with monitor.on_build_all(self):
        for project in self.children:
            if isinstance(project, Project):
                project.build()
            else:
                project.build_all()

    def publish_all(self, repository: Repository, auth: Authentication):
        """
        recursively run the publish operation on all projects and subprojects in this group
        :param repository: the repository to publish into
        :param auth: authentication for this repository
        """

        # with monitor.on_publish_all(self):
        for project in self.children:
            if isinstance(project, Project):
                project.publish(repository, auth)
            else:
                project.publish_all(repository, auth)

    def install_all(self):
        """
        recursively run the 'install' (with dependencies) operation on all projects and subprojects in this group
        """
        # with monitor.on_install_all(self):
        for project in self.children:
            if isinstance(project, Project):
                project.dev_install()
            else:
                project.install_all()

    @classmethod
    def _find_parent(cls, path: Path) -> Optional["ProjectGroup"]:
        for path_parent in path.parents:
            if (parent_config_file := (path_parent / 'pyproject-group.toml')).exists():
                parent_config = PyProjectGroupConfiguration.load(parent_config_file)
                if any(child == path for child in parent_config.children):
                    return ProjectGroup(parent_config)
        return None

    @classmethod
    def of(cls, project: Project) -> Optional["ProjectGroup"]:
        """
        :param project: the project to get the project group for
        :return: the project group if such defined
        """
        if group := project.config.pkm_project.group:
            return ProjectGroup(PyProjectGroupConfiguration.load(resolve_relativity(Path(group), project.path)))
        return cls._find_parent(project.path)

    @classmethod
    def load(cls, path: Path) -> "ProjectGroup":
        """
        load the project group from a specific path
        :param path: the path to the project group directory
        :return: the loaded project group
        """
        return ProjectGroup(PyProjectGroupConfiguration.load(path / 'pyproject-group.toml'))

    @staticmethod
    def is_valid(path: Path) -> bool:
        """
        :param path: the path to check
        :return: True if the path contain project group, False otherwise
        """
        return (path / 'pyproject-group.toml').exists()


class PyProjectGroupConfiguration(TomlFileConfiguration):

    @computed_based_on("name")
    def name(self) -> str:
        return self['name'] or self._path.parent.name

    @property
    def path(self) -> Path:
        return self._path

    @computed_based_on("project-group.children")
    def children(self) -> Tuple[Path, ...]:
        if not (children := self["project-group.children"]):
            return tuple()

        root = self.path.parent
        result = []
        for path in children:
            path = Path(path)
            if path.is_absolute():
                result.append(path.resolve())
            else:
                result.append((root / path).resolve())

        return tuple(result)

    @children.modifier
    def set_children(self, new: Iterable[Path]):
        root = self.path.parent
        result = []
        for path in new:
            if path.is_absolute():
                result.append(str(path_to(root, path)))
            else:
                result.append(str(path))

        self["project-group.children"] = result

    @computed_based_on("project-group.parent")
    def parent(self) -> Optional[Path]:
        if parent := self["project-group.parent"]:
            path = Path(parent)
            if path.is_absolute():
                return path.resolve()
            return (self.path.parent / path).resolve()
        return None



from __future__ import annotations
import re
from dataclasses import dataclass, replace
from pathlib import Path
from typing import List, Optional, Union, Dict, Mapping, Any, Iterator

from pkm.api.dependencies.dependency import Dependency
from pkm.api.dependencies.env_markers import EnvironmentMarker
from pkm.api.distributions.distinfo import EntryPoint, ObjectReference
from pkm.api.packages.package import PackageDescriptor
from pkm.api.versions.version import Version
from pkm.api.versions.version_specifiers import VersionSpecifier, StandardVersionRange, AllowAllVersions
from pkm.config.configuration import TomlFileConfiguration, computed_based_on
from pkm.resolution.pubgrub import MalformedPackageException
from pkm.utils.commons import unone
from pkm.utils.dicts import remove_none_values, without_keys
from pkm.utils.files import path_to
from pkm.utils.properties import cached_property
from pkm.utils.sequences import strs


@dataclass(frozen=True, eq=True)
class BuildSystemConfig:
    requirements: List[Dependency]
    build_backend: str
    backend_path: Optional[List[str]]

    def to_config(self) -> Dict[str, Any]:
        return remove_none_values({
            'requires': [str(d) for d in self.requirements] if self.requirements else None,
            'build-backend': self.build_backend,
            'backend-path': self.backend_path
        })

    @classmethod
    def from_config(cls, cfg: Dict[str, Any]) -> "BuildSystemConfig":
        requirements = [Dependency.parse(dep) for dep in (cfg.get('requires') or [])]
        build_backend = cfg.get('build-backend')
        backend_path = cfg.get('backend-path')

        return BuildSystemConfig(requirements, build_backend, backend_path)


@dataclass(frozen=True, eq=True)
class ContactInfo:
    name: Optional[str] = None
    email: Optional[str] = None

    @classmethod
    def from_config(cls, contact: Dict[str, Any]) -> "ContactInfo":
        return cls(**contact)

    def to_config(self) -> Dict[str, Any]:
        return remove_none_values({
            'name': self.name, 'email': self.email
        })


def _entrypoints_from_config(group: str, ep: Dict[str, str]) -> List[EntryPoint]:
    return [EntryPoint(group, ep_name, ObjectReference.parse(ep_oref)) for ep_name, ep_oref in ep.items()]


def _entrypoints_to_config(entries: List[EntryPoint]) -> Dict[str, str]:
    return {e.name: str(e.ref) for e in entries}


PKM_DIST_CFG_TYPE_LIB = "lib"
PKM_DIST_CFG_TYPE_CAPP = "cnt-app"


@dataclass(frozen=True, eq=True)
class PkmDistributionConfig:
    type: str  # lib, cnt-app

    def to_config(self) -> Dict[str, Any]:
        return {**self.__dict__}

    @classmethod
    def from_config(cls, cfg: Dict[str, Any]) -> PkmDistributionConfig:
        return cls(**cfg)


@dataclass(frozen=True, eq=True)
class PkmApplicationConfig:
    containerized: bool

    dependencies: List[Dependency]
    dependency_overwrites: Dict[str, Dependency]
    exposed_packages: List[str]

    def to_config(self) -> Dict[str, Any]:
        return {
            'containerized': self.containerized,
            'dependencies': strs(self.dependencies),
            'dependency-overwrites': {p: str(d) for p, d in self.dependency_overwrites.items()},
            'exposed-packages': self.exposed_packages
        }

    @classmethod
    def from_config(cls, config: Dict[str, Any]):
        return cls(
            config.get('containerized', False),
            [Dependency.parse(it) for it in unone(config.get('dependencies'), list)],
            {pk: Dependency.parse(dep) for pk, dep in unone(config.get('dependency-overwrites'), dict).items()},
            unone(config.get('exposed-packages'), list)
        )


@dataclass(frozen=True, eq=True)
class PkmProjectConfig:
    packages: Optional[List[str]] = None
    group: Optional[str] = None  # TODO: remove group and use conventions only?

    def to_config(self) -> Dict[str, Any]:
        return remove_none_values({
            'packages': self.packages,
            'group': self.group
        })

    @classmethod
    def from_config(cls, config: Optional[Dict[str, Any]]) -> Optional["PkmProjectConfig"]:
        if not config:
            return PkmProjectConfig()

        return PkmProjectConfig(**config)


@dataclass(frozen=True, eq=True)
class ProjectConfig:
    """
    the project config as described in
    https://www.python.org/dev/peps/pep-0621/, https://www.python.org/dev/peps/pep-0631/
    """

    # The name of the project.
    name: str
    # The version of the project as supported by PEP 440.
    version: Version
    # The summary description of the project.
    description: Optional[str]
    # The actual text or Path to a text file containing the full description of this project.
    readme: Union[Path, str, None]
    # The Python version requirements of the project.
    requires_python: Optional[Union[StandardVersionRange, AllowAllVersions]]
    # The project licence identifier or path to the actual licence file
    license: Union[str, Path, None]
    # The people or organizations considered to be the "authors" of the project.
    authors: Optional[List[ContactInfo]]
    # similar to "authors", exact meaning is open to interpretation.
    maintainers: Optional[List[ContactInfo]]
    # The keywords for the project.
    keywords: Optional[List[str]]
    # Trove classifiers (https://pypi.org/classifiers/) which apply to the project.
    classifiers: Optional[List[str]]
    # A mapping of URLs where the key is the URL label and the value is the URL itself.
    urls: Optional[Dict[str, str]]
    # list of entry points, following https://packaging.python.org/en/latest/specifications/entry-points/.
    entry_points: Optional[Dict[str, List[EntryPoint]]]
    # The dependencies of the project.
    dependencies: Optional[List[Dependency]]
    # The optional dependencies of the project, grouped by the 'extra' name that provides them.
    optional_dependencies: Dict[str, List[Dependency]]
    # a list of field names (from the above fields), each field name that appears in this list means that the absense of
    # data in the corresponding field means that a user tool provides it dynamically
    dynamic: Optional[List[str]]

    all_fields: Dict[str, Any]

    def all_entrypoints(self) -> List[EntryPoint]:
        if not self.entry_points:
            return []

        return [e for points in self.entry_points.values() for e in points]

    def is_dynamic(self, field: str) -> bool:
        """
        :param field: the field name (as in the configuration, e.g.,
                      optional-dependencies and not optional_dependencies)
        :return: True if the field is marked as dynamic, False otherwise
        """
        return self.all_fields.get(field) is None and bool(d := self.dynamic) and field in d  # noqa

    @cached_property
    def all_dependencies(self) -> List[Dependency]:
        all_deps = [d for d in (self.dependencies or [])]
        optional_deps = self.optional_dependencies or {}
        for od_group, deps in optional_deps.items():
            extra_rx = re.compile(f'extra\\s*==\\s*(\'{od_group}\'|"{od_group}")')
            for dep in deps:
                if not dep.env_marker or not extra_rx.match(str(dep.env_marker)):
                    new_marker = (str(dep.env_marker).rstrip(';') + ';') if dep.env_marker else ''
                    new_marker = f"{new_marker}extra==\'{od_group}\'"
                    all_deps.append(replace(dep, env_marker=EnvironmentMarker.parse_pep508(new_marker)))
                else:
                    all_deps.append(dep)
        return all_deps

    def package_descriptor(self) -> PackageDescriptor:
        return PackageDescriptor(self.name, self.version)

    def readme_content(self) -> str:
        if not self.readme:
            return ""

        if isinstance(self.readme, str):
            return self.readme

        if self.readme.exists():
            return self.readme.read_text()

        return ""

    def script_entrypoints(self) -> Iterator[EntryPoint]:
        """
        :return: iterator over the scripts (and gui scripts) entrypoints
        """
        if scripts := self.entry_points.get(EntryPoint.G_CONSOLE_SCRIPTS):
            yield from scripts

        if gui_scripts := self.entry_points.get(EntryPoint.G_GUI_SCRIPTS):
            yield from gui_scripts

    def readme_content_type(self) -> str:
        if self.readme and isinstance(self.readme, Path):
            readme_suffix = self.readme.suffix
            if readme_suffix == '.md':
                return 'text/markdown'
            elif readme_suffix == '.rst':
                return 'text/x-rst'

        return 'text/plain'

    def license_content(self) -> str:
        if not self.license:
            return ""

        if isinstance(self.license, str):
            return self.license

        return self.license.read_text()

    def to_config(self, project_path: Optional[Path] = None) -> Dict[str, Any]:
        project_path = project_path or Path.cwd()
        readme_value = None
        if self.readme:
            readme_value = self.readme if isinstance(self.readme, str) \
                else {'file': str(path_to(project_path, self.readme))}

        ep: Dict[str, List[EntryPoint]] = self.entry_points or {}
        ep_no_scripts: Dict[str, List[EntryPoint]] = without_keys(
            ep, EntryPoint.G_CONSOLE_SCRIPTS, EntryPoint.G_GUI_SCRIPTS)

        optional_dependencies = {
            extra: [str(d) for d in deps]
            for extra, deps in self.optional_dependencies.items()
        } if self.optional_dependencies else None

        license_ = None
        if self.license:
            license_ = {'file': self.license} if isinstance(self.license, Path) else {'text': self.license}

        project = {
            **self.all_fields,
            'name': self.name, 'version': str(self.version), 'description': self.description,
            'readme': readme_value, 'requires-python': str(self.requires_python) if self.requires_python else None,
            'license': license_, 'authors': [c.to_config() for c in self.authors] if self.authors is not None else None,
            'maintainers': [c.to_config() for c in self.maintainers] if self.maintainers is not None else None,
            'keywords': self.keywords, 'classifiers': self.classifiers, 'urls': self.urls,
            'scripts': _entrypoints_to_config(
                ep[EntryPoint.G_CONSOLE_SCRIPTS]) if EntryPoint.G_CONSOLE_SCRIPTS in ep else None,
            'gui-scripts': _entrypoints_to_config(
                ep[EntryPoint.G_GUI_SCRIPTS]) if EntryPoint.G_GUI_SCRIPTS in ep else None,
            'entry-points': {group: _entrypoints_to_config(entries)
                             for group, entries in ep_no_scripts.items()} if ep_no_scripts else None,
            'dependencies': [str(d) for d in self.dependencies] if self.dependencies else None,
            'optional-dependencies': optional_dependencies, 'dynamic': self.dynamic
        }

        return remove_none_values(project)

    @classmethod
    def from_config(cls, config: Dict[str, Any], project_path: Optional[Path] = None) -> "ProjectConfig":
        project_path = project_path or Path.cwd()

        version = Version.parse(config['version'])

        # decide the readme value
        readme = None
        if readme_entry := config.get('readme'):
            if isinstance(readme_entry, Mapping):
                if readme_file := readme_entry.get('file'):
                    readme = project_path / readme_file
                else:
                    readme = str(readme_entry['text'])
            else:
                readme = str(readme_entry)

        requires_python = VersionSpecifier.parse(config['requires-python']) \
            if 'requires-python' in config else AllowAllVersions

        license_ = None
        if license_table := config.get('license'):
            license_ = (project_path / license_table['file']) if 'file' in license_table else str(license_table['text'])

        authors = None
        if authors_array := config.get('authors'):
            authors = [ContactInfo.from_config(a) for a in authors_array]

        maintainers = None
        if maintainers_array := config.get('maintainers'):
            maintainers = [ContactInfo.from_config(a) for a in maintainers_array]

        entry_points = {}
        if scripts_table := config.get('scripts'):
            entry_points[EntryPoint.G_CONSOLE_SCRIPTS] = _entrypoints_from_config(
                EntryPoint.G_CONSOLE_SCRIPTS, scripts_table)

        if gui_scripts_table := config.get('gui-scripts'):
            entry_points[EntryPoint.G_GUI_SCRIPTS] = _entrypoints_from_config(
                EntryPoint.G_GUI_SCRIPTS, gui_scripts_table)

        if entry_points_tables := config.get('entry-points'):
            entry_points.update({
                group: _entrypoints_from_config(group, entries)
                for group, entries in entry_points_tables.items()
            })

        dependencies = None
        if dependencies_array := config.get('dependencies'):
            dependencies = [Dependency.parse(it) for it in dependencies_array]

        optional_dependencies = {}
        if optional_dependencies_table := config.get('optional-dependencies'):
            optional_dependencies = {
                extra: [Dependency.parse(it) for it in deps]
                for extra, deps in optional_dependencies_table.items()
            }

        return ProjectConfig(
            name=config['name'], version=version, description=config.get('description'), readme=readme,
            requires_python=requires_python, license=license_, authors=authors, maintainers=maintainers,
            keywords=config.get('keywords'), classifiers=config.get('classifiers'), urls=config.get('urls'),
            entry_points=entry_points, dependencies=dependencies, optional_dependencies=optional_dependencies,
            dynamic=config.get('dynamic'), all_fields=config)


_LEGACY_BUILDSYS = {
    'requires': ['setuptools', 'wheel', 'pip', 'cython'],
    'build-backend': 'setuptools.build_meta:__legacy__'
}

_LEGACY_PROJECT = {
    'dynamic': [
        'description', 'readme', 'requires-python', 'license', 'authors', 'maintainers', 'keywords',
        'classifiers', 'urls', 'scripts', 'gui-scripts', 'entry-points', 'dependencies', 'optional-dependencies']}


class PyProjectConfiguration(TomlFileConfiguration):
    # here due to pycharm bug https://youtrack.jetbrains.com/issue/PY-47698
    project: ProjectConfig
    pkm_project: PkmProjectConfig
    build_system: BuildSystemConfig
    pkm_application: PkmApplicationConfig
    pkm_distribution: PkmDistributionConfig

    @computed_based_on("tool.pkm.project")
    def pkm_project(self) -> PkmProjectConfig:
        return PkmProjectConfig.from_config(self['tool.pkm.project'])

    @computed_based_on("tool.pkm.application")
    def pkm_application(self) -> Optional[PkmApplicationConfig]:
        if app_config := self['tool.pkm.application']:
            return PkmApplicationConfig.from_config(app_config)
        return None

    @pkm_application.modifier
    def set_pkm_application(self, app: Optional[PkmApplicationConfig]):
        if app is None:
            del self['tool.pkm.application']
        else:
            self['tool.pkm.application'] = app.to_config()

    @computed_based_on("tool.pkm.distribution")
    def pkm_distribution(self) -> PkmDistributionConfig:
        if dist_config := self['tool.pkm.distribution']:
            return PkmDistributionConfig.from_config(dist_config)

        return PkmDistributionConfig(PKM_DIST_CFG_TYPE_LIB)

    @pkm_distribution.modifier
    def set_pkm_distribution(self, distribution: PkmDistributionConfig):
        self['tool.pkm.distribution'] = distribution.to_config()

    @computed_based_on("project")
    def project(self) -> Optional[ProjectConfig]:
        project_path = self._path.parent if self._path else None

        project: Dict[str, Any] = self['project']
        if project is None:
            return None

        return ProjectConfig.from_config(project, project_path)

    @project.modifier
    def set_project(self, value: ProjectConfig):
        project_path = self.path and self.path.parent
        self['project'] = value.to_config(project_path)

    @computed_based_on("build-system")
    def build_system(self) -> BuildSystemConfig:
        return BuildSystemConfig.from_config(self['build-system'])

    @build_system.modifier
    def set_build_system(self, bs: BuildSystemConfig):
        self['build-system'] = bs.to_config()

    @classmethod
    def load_effective(cls, pyproject_file: Path,
                       package: Optional[PackageDescriptor] = None) -> "PyProjectConfiguration":
        """
        load the effective pyproject file (with missing essential values filled with their legacy values)
        for example, if no build-system is available, this method will fill in the legacy build-system
        :param pyproject_file: the pyproject.toml to load
        :param package: the package that this pyproject belongs to, if given,
                        it will be used in case of missing name and version values
        :return: the loaded pyproject
        """
        pyproject = PyProjectConfiguration.load(pyproject_file)
        source_tree = pyproject_file.parent

        # ensure build-system:
        if pyproject['build-system'] is None:
            if not (source_tree / 'setup.py').exists():
                package_info = str(pyproject_file.parent) if not package else f"{package.name} {package.version}"
                raise MalformedPackageException(f"cannot infer build system for package: {package_info}")

            pyproject['build-system'] = _LEGACY_BUILDSYS

        if pyproject['build-system.requires'] is None:
            pyproject['build-system.requires'] = []

        if pyproject['build-system.build-backend'] is None:
            pyproject['build-system.build-backend'] = _LEGACY_BUILDSYS['build-backend']
            pyproject['build-system.requires'] = list(
                {*_LEGACY_BUILDSYS['requires'], *pyproject['build-system.requires']})

        # ensure project:
        if not pyproject['project']:
            pyproject['project'] = {
                **_LEGACY_PROJECT,
                'name': (package or source_tree).name,
                'version': str(package.version) if package else 'unknown_version'}

        pyproject['project.name'] = PackageDescriptor.normalize_name(pyproject['project.name'])
        return pyproject



from __future__ import annotations

import warnings
from collections import defaultdict
from dataclasses import dataclass, field
from pathlib import Path
from typing import List, Dict, Iterable

from pkm.api.dependencies.dependency import Dependency
from pkm.api.environments.environment import Environment
from pkm.api.packages.package import Package
from pkm.api.repositories.repositories_configuration import RepositoryInstanceConfig, RepositoriesConfiguration, \
    RepositoriesConfigInheritanceMode
from pkm.api.repositories.repository import Repository, RepositoryBuilder, AbstractRepository
from pkm.repositories.file_repository import FileRepository
from pkm.utils.dicts import put_if_absent
from pkm.utils.http.http_client import HttpClient

REPOSITORIES_EXTENSIONS_ENTRYPOINT_GROUP = "pkm-repositories"
REPOSITORIES_CONFIGURATION_PATH = "etc/pkm/repositories.toml"


class RepositoryLoader:
    def __init__(self, main_cfg: Path, http: HttpClient, workspace: Path):

        from pkm.api.environments.environment import Environment
        from pkm.repositories.simple_repository import SimpleRepositoryBuilder
        from pkm.repositories.git_repository import GitRepository
        from pkm.repositories.pypi_repository import PyPiRepository
        from pkm.repositories.local_packages_repository import LocalPackagesRepositoryBuilder
        from pkm.repositories.url_repository import UrlRepository

        # common builders
        self._builders = {
            b.name: b for b in (
                SimpleRepositoryBuilder(http),
                LocalPackagesRepositoryBuilder(),
            )
        }

        self.pypi = PyPiRepository(http)
        base_repositories = [
            self.pypi, GitRepository(workspace / 'git'), UrlRepository(), FileRepository()]

        # builders from entrypoints
        for epoint in Environment.current().entrypoints[REPOSITORIES_EXTENSIONS_ENTRYPOINT_GROUP]:
            try:
                ext: RepositoriesExtension = epoint.ref.import_object()()
                if not isinstance(ext, RepositoriesExtension):
                    raise ValueError("repositories entrypoint did not provide to a RepositoriesExtension class")

                for builder in ext.builders:
                    self._builders[builder.name] = builder

                base_repositories.extend(ext.instances)

            except Exception:  # noqa
                import traceback
                warnings.warn(f"malformed repository entrypoint: {epoint}")
                traceback.print_exc()

        base = _CompositeRepository('base', base_repositories, {})

        self._cached_instances: Dict[RepositoryInstanceConfig, Repository] = {}  # noqa

        self._main = self.load('main', RepositoriesConfiguration.load(main_cfg), base)
        self.workspace = workspace

    @property
    def main(self) -> Repository:
        return self._main

    def load(self, name: str, config: RepositoriesConfiguration, next_in_context: Repository) -> Repository:
        package_search_list = []
        package_associated_repo = {}

        if not (new_repos := config.repositories):
            if config.inheritance_mode == RepositoriesConfigInheritanceMode.INHERIT_MAIN:
                return self.main
            return next_in_context

        for definition in new_repos:
            print(f"DBG: building instance by definition {definition}")
            instance = self.build(definition)

            if definition.packages:
                for package in definition.packages:
                    if package == '*':
                        package_search_list.append(instance)
                    else:
                        put_if_absent(package_associated_repo, package, instance)
            else:
                package_search_list.append(instance)

        if config.inheritance_mode == RepositoriesConfigInheritanceMode.INHERIT_CONTEXT:
            package_search_list.append(next_in_context)
        elif config.inheritance_mode == RepositoriesConfigInheritanceMode.INHERIT_MAIN:
            package_search_list.append(self.main)

        return _CompositeRepository(name, package_search_list, package_associated_repo)

    def build(self, config: RepositoryInstanceConfig) -> Repository:
        if not (cached := self._cached_instances.get(config)):
            if not (builder := self._builders.get(config.type)):
                raise KeyError(f"unknown repository type: {config.type}")
            cached = builder.build(config.name, config.packages, **config.args)
            self._cached_instances[config] = cached

        return cached


class _CompositeRepository(AbstractRepository):
    def __init__(
            self, name: str, package_search_list: List[Repository],
            package_associated_repos: Dict[str, Repository]):
        super().__init__(name)

        self._url_handlers: Dict[str, List[Repository]] = defaultdict(list)
        self._package_search_list = []

        for repo in package_search_list:
            if repo.accept_non_url_packages():
                self._package_search_list.append(repo)
            for protocol in repo.accepted_url_protocols():
                self._url_handlers[protocol].append(repo)

        self._package_associated_repos = package_associated_repos

    def _do_match(self, dependency: Dependency, env: Environment) -> List[Package]:

        if url := dependency.required_url():
            for repo in self._url_handlers[url.protocol]:
                if match := repo.match(dependency, env):
                    return match
            raise []

        if repo := self._package_associated_repos.get(dependency.package_name):
            return repo.match(dependency, env)

        for repo in self._package_search_list:
            if result := repo.match(dependency, env):
                return result

        return []

    def _sort_by_priority(self, dependency: Dependency, packages: List[Package]) -> List[Package]:
        return packages

    def accepted_url_protocols(self) -> Iterable[str]:
        return self._url_handlers.keys()


@dataclass
class RepositoriesExtension:
    builders: List[RepositoryBuilder] = field(default_factory=list)
    instances: List[Repository] = field(default_factory=list)



from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, Any

from pkm.api.versions.version import Version
from pkm.utils.monitors import MonitoredOperation, MonitoredEvent
from pkm.utils.types import IterableWithLen


@dataclass
class DependencyResolutionMonitoredOp(MonitoredOperation):
    ...


@dataclass
class DependencyResolutionIterationEvent(MonitoredEvent):
    packages_completed: IterableWithLen[str]
    packages_requested: IterableWithLen[str]
    current_package: Any


@dataclass
class DependencyResolutionConclusionEvent(MonitoredEvent):
    decisions: Dict[Any, Version]



import re
from copy import copy
from dataclasses import dataclass
from numbers import Number
from pathlib import Path
from typing import Optional, Dict, Any, Mapping, List, Tuple, Union, MutableMapping, Callable

from pkm.utils.parsers import SimpleParser

_KEY_RX = re.compile(r'[a-zA-Z\d\-_]+')
KEY_T = Tuple
DUMPS_T = Callable[[Dict[str, Any]], str]


class TimeLiteral(str):
    pass


class Table(Dict[str, Any]):

    def __init__(self, *args, style: Optional[Dict[str, str]] = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.style: Dict[str, str] = style if style is not None else {'display': 'auto'}


class Array(List[Any]):
    def __init__(self, *args, style: Optional[Dict[str, str]] = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.style: Dict[str, str] = style if style is not None else {'display': 'auto'}


@dataclass
class _ValueStyle:
    prolog: str = ''
    epilog: str = ''
    content_trailing: str = ''

    value_str: Optional[str] = None
    value_sig: int = 0

    display: str = 'auto'


@dataclass
class _KeyStyle:
    prolog: str = ''
    epilog: str = ''
    key_str: Optional[str] = None
    display: str = 'auto'


_DEFAULT_VALUE_STYLE = _ValueStyle()
_DEFAULT_KEY_STYLE = _KeyStyle()


class _StyleSheet:
    def __init__(self):
        self.keys: Dict[KEY_T, _KeyStyle] = {}
        self.values: Dict[KEY_T, _ValueStyle] = {}

    def lastify(self, key: KEY_T) -> KEY_T:
        lk = len(key) + 1
        index = sum(len(k) == lk and k[:-1] == key for k in self.keys.keys())
        return *key, index

    def set_key_style(self, key: KEY_T, style: _KeyStyle):
        self.keys[key] = style

    def set_value_style(self, key: KEY_T, value: Any, style: _ValueStyle):
        if isinstance(value, (Mapping, List)):
            style.value_sig = type(value)
        else:
            style.value_sig = id(value)

        self.values[key] = style

    def update_value_auto_display(self, key: KEY_T, value: Any, display: str):
        if key not in self.values:
            self.set_value_style(key, value, _ValueStyle(display=display))
        else:
            vstyle = self.values[key]
            if vstyle.display == 'auto':
                vstyle.display = display

    def key_style(self, key: KEY_T, display: str = 'auto') -> _KeyStyle:
        result = self.keys.get(key) or _DEFAULT_KEY_STYLE

        if display != 'auto' and result.display != display:
            result = copy(result)
            result.display = display
            if display == 'assignment':
                result.epilog = ' '

        return result

    def value_style(self, key: KEY_T, value: Any, display: str = 'auto') -> _ValueStyle:
        predefined_style = self.values.get(key)
        if not predefined_style or \
                (predefined_style.value_sig is not type(value) and predefined_style.value_sig != id(value)):
            predefined_style = _DEFAULT_VALUE_STYLE

        if isinstance(value, (Table, Array)):
            predefined_style = _ValueStyle(**value.style)

        if predefined_style.display != 'auto' and (display == 'auto' or display == predefined_style.display):
            return predefined_style

        vstyle = copy(predefined_style)

        if display != 'auto':
            vstyle.display = display
        else:
            vstyle.display = 'inline'
            if isinstance(value, Mapping):
                is_root = not key
                is_first_level = len(key) == 1
                has_no_assignments = all(
                    isinstance(it, Mapping) or
                    (isinstance(it, List) and self.value_style((*key, k), it).display == 'table-array')
                    for k, it in value.items())
                contains_large_data = \
                    len(value) > 5 or \
                    any(isinstance(it, (Mapping, List)) or (isinstance(it, str) and len(it) > 40)
                        for it in value.values())

                if is_root:
                    vstyle.display = 'regular'
                elif has_no_assignments:
                    vstyle.display = 'hidden'
                elif contains_large_data or is_first_level:
                    vstyle.display = 'regular'

            elif isinstance(value, List):
                has_items = len(value) > 0
                has_no_assignments = all(isinstance(it, Mapping) for it in value)
                if has_items and has_no_assignments:
                    vstyle.display = 'table-array'

        # configure auto display
        if vstyle.display == 'inline':
            vstyle.prolog = ' '

        return vstyle


# noinspection PyMethodMayBeStatic
class _Writer:
    def __init__(self, style: Optional[_StyleSheet] = None):
        self.style = style or _StyleSheet()

    def write(self, data: Dict[str, Any]) -> str:
        key = ()
        return self._write_value(data, key, self.style.value_style(key, data)).rstrip()

    def _write_value(self, data: Any, key: KEY_T, vstyle: _ValueStyle) -> str:
        if isinstance(data, bool):
            return self._write_bool(data, vstyle)

        if isinstance(data, (Number, TimeLiteral)):
            return self._write_nummeric_like(data, vstyle)

        if isinstance(data, str):
            return self._write_str(data, vstyle)

        if isinstance(data, Mapping):
            return self._write_table(data, key, vstyle)

        if isinstance(data, List):
            return self._write_list(data, key, vstyle)

        raise ValueError(f"Unknown value type: {type(data)} for key {key}")

    def _write_nummeric_like(self, data: Any, style: _ValueStyle) -> str:
        v = style.value_str
        if v is None:
            v = str(data)

        return f"{style.prolog}{v}{style.epilog}"

    def _write_bool(self, data: Any, style: _ValueStyle) -> str:
        v = style.value_str
        if v is None:
            v = 'true' if data else 'false'

        return f"{style.prolog}{v}{style.epilog}"

    def _write_list(self, data: List[Any], key: KEY_T, vstyle: _ValueStyle) -> str:
        if vstyle.display == 'inline':  # TODO: support inline column and inline grid
            return self._write_inline_list(data, key, vstyle)
        return self._write_table_list(data, key, vstyle)

    def _write_inline_list(self, data: List[Any], key: KEY_T, vstyle: _ValueStyle) -> str:
        result = f"{vstyle.prolog}["

        def _write_value(v: Any, i: int) -> str:
            k = (*key, i)
            s = self.style.value_style(k, v, 'inline')
            return self._write_value(v, k, s)

        result += ', '.join(_write_value(v, i) for i, v in enumerate(data))
        result += f'{vstyle.content_trailing}]{vstyle.epilog}'
        return result

    def _write_table(self, data: Mapping[str, Any], key: KEY_T, vstyle: _ValueStyle) -> str:
        if vstyle.display == 'regular':
            return self._write_regular_table(data, key, vstyle)

        if vstyle.display == 'hidden':
            return self._write_hidden_table(data, key, vstyle)

        if vstyle.display == 'inline':
            return self._write_inline_table(data, key, vstyle)

        if vstyle.display == 'dotted':
            return self._write_dotted_table(data, key, vstyle)

        raise ValueError(f"unknown table display style: {vstyle.display}")

    def _write_str(self, data: str, vstyle: _ValueStyle) -> str:
        value_str = vstyle.value_str
        if value_str is None:
            escaped = data.translate({
                ord('\b'): '\\b',
                ord('\t'): '\\t',
                ord('\n'): '\\n',
                ord('\f'): '\\f',
                ord('\r'): '\\r',
                ord('"'): '\\"',
                ord('\\'): '\\\\',
            })

            value_str = f'"{escaped}"'

        return f'{vstyle.prolog}{value_str}{vstyle.epilog}'

    def _write_key(self, key: KEY_T, offset: int, display: str) -> str:

        if not key:
            return ''

        key_style = self.style.key_style(key, display)
        result = key_style.prolog

        key_str = key_style.key_str
        if not key_str:
            key_str = ''
            if display == 'table':
                key_str += '['
            elif display == 'table-array-item':
                key_str += '[['

            key_str += '.'.join(
                self._write_key_part(key[0:i])
                for i in range(offset + 1, len(key) + 1)
                if isinstance(key[i - 1], str))

            if display == 'table':
                key_str += ']'
            elif display == 'table-array-item':
                key_str += ']]'

        result += key_str
        result += key_style.epilog
        if display != 'assignment':
            result += '\n'

        return result

    def _write_key_part(self, part: KEY_T) -> str:
        if _KEY_RX.fullmatch(part[-1]):
            return part[-1]
        else:
            return self._write_str(part[-1], _DEFAULT_VALUE_STYLE)

    def _write_regular_table(
            self, data: Mapping[str, Any],
            key: KEY_T, vstyle: _ValueStyle) -> str:

        result = vstyle.prolog
        key_display = vstyle.display if vstyle.display == 'table-array-item' else 'table'
        result += self._write_key(key, 0, key_display)

        delayed_items: Dict[KEY_T, Any] = {}
        has_assignments = False

        for partial_key, item in data.items():
            item_key = (*key, partial_key)
            item_style = self.style.value_style(item_key, item)

            if item_style.display == 'inline':
                result += f"{self._write_key(item_key, len(key), 'assignment')}" \
                          f"={self._write_value(item, item_key, item_style)}\n"
                has_assignments = True
            else:
                if item_style.display == 'dotted':
                    result += self._write_value(item, item_key, self.style.value_style(item_key, item))
                    has_assignments = True
                else:
                    delayed_items[item_key] = item

        if has_assignments:
            result += '\n'

        for item_key, item in delayed_items.items():
            result += self._write_value(item, item_key, self.style.value_style(item_key, item))

        result += vstyle.epilog

        return result

    # noinspection PyUnusedLocal
    def _write_hidden_table(self, data: Mapping[str, Any], key: KEY_T, vstyle: _ValueStyle) -> str:
        result = ''

        for partial_key, item in data.items():
            item_key = (*key, partial_key)

            is_list = isinstance(item, List)
            is_table = isinstance(item, Mapping)

            inappropriate_item = not is_list and not is_table
            inappropriate_item = inappropriate_item or is_list and any(not isinstance(it, Mapping) for it in item)

            if inappropriate_item:
                return self._write_regular_table(data, key, self.style.value_style(key, data, 'regular'))

            item_style = self.style.value_style(item_key, item)
            if item_style.display == 'inline':
                item_style = self.style.value_style(item_key, item, 'regular')

            if is_list:
                result += self._write_table_list(item, item_key, item_style)
            elif item_style.display == 'hidden':
                result += self._write_hidden_table(item, item_key, item_style)
            else:
                result += self._write_regular_table(item, item_key, item_style)

        return result

    def _write_inline_table(self, data: Mapping[str, Any], key: KEY_T, vstyle: _ValueStyle) -> str:
        result = f"{vstyle.prolog}{{"

        def _write_assignment(partial_key: str, item: Any) -> str:
            item_key = (*key, partial_key)
            item_style = self.style.value_style(item_key, item, 'inline')

            return f"{self._write_key(item_key, len(key), 'assignment')}" \
                   f"={self._write_value(item, item_key, item_style)}"

        result += ', '.join(_write_assignment(k, v) for k, v in data.items())
        result += f"}}{vstyle.epilog}"
        return result

    def _write_dotted_table(self, data: Mapping[str, Any], key: KEY_T, vstyle: _ValueStyle) -> str:
        subkey = self._write_key_part(key)
        result = ""

        def _write_assignment(partial_key: str, item: Any) -> str:
            item_key = (*key, partial_key)
            key_style = self.style.key_style(item_key)
            key_str = self._write_key(item_key, len(key), 'assignment')
            if not key_style.key_str:
                key_str = f"{subkey}.{key_str}"
            item_style = self.style.value_style(item_key, item, 'inline')
            value_str = self._write_value(item, item_key, item_style)

            if isinstance(item, Mapping) and item_style.display == 'dotted':
                return value_str
            return f"{key_str}={value_str}"

        result += '\n'.join(_write_assignment(k, v) for k, v in data.items())
        result += '\n'
        result += vstyle.epilog
        return result

    def _write_table_list(self, data: List[Mapping[str, Any]], key: KEY_T, vstyle: _ValueStyle) -> str:
        def _write_table_item(table: Mapping[str, Any], index: int):
            table_key = (*key, index)
            table_style = self.style.value_style(table_key, table, 'table-array-item')
            return self._write_regular_table(table, table_key, table_style)

        items = ''.join(_write_table_item(it, i) for i, it in enumerate(data))
        return f"{vstyle.prolog}{items}{vstyle.epilog}"


class _TomlParser(SimpleParser):
    def __init__(self, text: str, file_name: Optional[str]):
        super().__init__(text, file_name)
        self.style = _StyleSheet()

    def read_non_data(self, allow_new_lines: bool = True) -> str:
        result = ''
        while self.is_not_empty():
            result += self.match_ws(allow_new_lines)
            n = self.peek()
            if n == '#':  # comment
                result += self.until_match('\n')
            else:
                break

        return result

    def read_multiline_basic_str(self) -> str:
        self.match_or_err('"""', 'multiline basic string expected')
        s = '"""' + self.until_match('"""')
        while s[-1] == '\\':
            s += self.next(3) + self.until_match('"""')

        s += '"""'
        self.match_or_err('"""', 'multiline basic string ending expected')

        return eval(s)

    def read_multiline_literal_str(self) -> str:
        self.match_or_err("'''", 'multiline literal string expected')
        string = self.until_match("'''")
        self.match_or_err("'''", "missing multiline literal string closing")
        return string

    def read_basic_str(self) -> str:
        self.match_or_err('"', 'basic string expected')
        s = '"' + self.until_match('"')
        while s[-1] == '\\':
            s += self.next() + self.until_match('"')

        s += '"'
        self.match_or_err('"', 'basic string ending expected')

        return eval(s)

    def read_literal_str(self) -> str:
        self.match_or_err("'", 'literal string expected')
        string = self.until_match("'")
        self.match_or_err("'", "missing literal string closing")
        return string

    def read_str(self) -> str:
        n = self.peek(3)
        value = ''

        if n == '"""':
            value = self.read_multiline_basic_str()
        elif n == "'''":
            value = self.read_multiline_literal_str()
        elif n[0] == '"':
            value = self.read_basic_str()
        elif n[0] == "'":
            value = self.read_literal_str()
        else:
            self.raise_err("string expected type")

        return value

    def read_bare_key(self) -> str:
        result = ''
        while self.is_not_empty():
            n = self.peek()
            if _KEY_RX.match(n):
                result += self.next()
            else:
                break

        return result

    def read_key(self, base_key: KEY_T = ()) -> KEY_T:
        prolog = self.read_non_data()
        p = self.position
        table_key = self.match('[')
        table_list_key = False

        if table_key:
            self.match_ws(False)
            table_list_key = self.match('[')

        parts = []
        while self.is_not_empty():
            self.match_ws(False)
            n = self.peek()
            if n in '"\'':
                parts.append(self.read_str())
            elif _KEY_RX.fullmatch(n):
                parts.append(self.read_bare_key())
            else:
                self.raise_err('non-key character')

            pp = self.position
            self.match_ws(False)
            if not self.match('.'):
                self.position = pp
                break

        key = tuple(parts)

        if table_key:
            self.match_or_err(']', 'expecting table key termination')

        if table_list_key:
            self.match_ws()
            self.match_or_err(']', 'expecting table list key termination')
            key = self.style.lastify(key)

        key_str = self.text[p:self.position]
        epilog = self.read_non_data(False)
        display = 'assignment'
        if table_key:
            display = 'table'
        if table_list_key:
            display = 'table-array-item'

        self.style.set_key_style((*base_key, *key),
                                 _KeyStyle(prolog=prolog, epilog=epilog, key_str=key_str, display=display))
        return key

    def _mark_internal_tables(self, table_key: KEY_T, key: KEY_T, display: str):
        for i in range(1, len(key) + 1):
            dtable_key = (*table_key, *key[0:i])
            self.style.update_value_auto_display(dtable_key, {}, display)

    def read_date(self) -> TimeLiteral:
        ts = self.until(lambda s, i: s[i].isnumeric() or s[i] in '-.: TZ')
        return TimeLiteral(ts)

    def read_number_or_date(self) -> Union[Number, TimeLiteral]:
        pos = self.position

        sign = (self.match('-') and '-') or (self.match('+') and '+') or ''
        nstr = sign + self.until(lambda i, s: not s[i].isnumeric() and s[i] not in '.e')

        n = self.peek()
        if n in '-:':
            self.position = pos
            return self.read_date()

        if not nstr:
            self.raise_err("number expected")

        return eval(nstr)

    def read_bool(self) -> bool:
        if self.match('true'):
            return True
        if self.match('false'):
            return False
        self.raise_err("boolean expected")

    def read_inline_list(self, key: KEY_T) -> List[Any]:
        prolog = self.read_non_data(False)
        self.match_or_err('[', 'expecting list start')
        result = []
        p = self.position
        content_trailing = None

        while self.is_not_empty():
            item_prolog = self.read_non_data()

            if self.match(']'):
                self.position -= 1
                content_trailing = self.text[p:self.position]
                break

            item_key = (*key, len(result))
            next_ = self.read_inline_value(item_key)
            item_style = self.style.value_style(item_key, next_)
            item_style.prolog = item_prolog
            item_style.epilog = self.read_non_data()
            result.append(next_)

            p = self.position
            if not self.match(', ') and not self.match(','):
                break

        content_trailing = content_trailing or self.read_non_data()
        self.match_or_err(']', 'expecting list end')
        epilog = self.read_non_data(False)
        self.style.set_value_style(key, result, _ValueStyle(prolog=prolog, epilog=epilog, display='inline',
                                                            content_trailing=content_trailing))
        return result

    def read_inline_table(self, key: KEY_T) -> Dict[str, Any]:
        prolog = self.read_non_data(False)
        self.match_or_err('{', 'expecting inline-table start')
        result = self.read_table_assignments(key, inline=True)
        self.match_or_err('}', 'expecting inline-table end')
        epilog = self.read_non_data(False)
        self.style.set_value_style(key, result, _ValueStyle(prolog=prolog, epilog=epilog, display='inline'))
        return result

    def read_inline_value(self, key: KEY_T) -> Any:
        prolog = self.read_non_data()
        value_str = None

        p = self.position
        n = self.peek()
        result = None
        if n in "+-" or n.isnumeric():
            result = self.read_number_or_date()
            value_str = self.text[p:self.position]
        elif n in '\'"':
            result = self.read_str()
            value_str = self.text[p:self.position]
        elif n in 'tf':
            result = self.read_bool()
            value_str = self.text[p:self.position]
        elif n == '[':
            result = self.read_inline_list(key)
        elif n == '{':
            result = self.read_inline_table(key)

        if result is None:
            self.raise_err("value expected")

        vstyle = self.style.value_style(key, result)
        vstyle.prolog = prolog
        vstyle.epilog = self.read_non_data(False)
        vstyle.value_str = value_str
        vstyle.display = 'inline'
        self.style.set_value_style(key, result, vstyle)

        return result

    def _enter(self, data: Any, key: KEY_T, value: Any):
        if not key:
            return data

        if isinstance(data, List):
            if len(key) == 1 and isinstance(key[0], Number):
                data.append(value)
                return

            return self._enter(data[-1], key, value)

        if isinstance(data, MutableMapping):
            if len(key) == 1:
                data[key[0]] = value
                return

            if key[0] not in data:
                if isinstance(key[1], Number):
                    data[key[0]] = []
                else:
                    data[key[0]] = {}

            return self._enter(data[key[0]], key[1:], value)

        raise ValueError(f"cannot enter {data}")

    def read_table_assignments(self, table_key: KEY_T, inline: bool = False) -> Dict[str, Any]:
        data = {}
        while self.is_not_empty():
            p = self.position
            self.read_non_data()
            key_start = self.peek()
            self.position = p
            if not key_start or key_start == '[':
                break

            k = self.read_key(base_key=table_key)
            self.match_or_err('=', 'expecting assignment')
            v = self.read_inline_value((*table_key, *k))
            self._enter(data, k, v)
            if len(k) > 1:
                self._mark_internal_tables(table_key, k[:-1], 'dotted')

            if inline and not self.match(', ') and not self.match(','):
                break
            if not inline and not self.match('\n'):
                break

        self.match('\n')  # drop last \n seperating assignments and next table - it is added implicitly
        return data

    def read_regular_table(self) -> Tuple[KEY_T, Dict[str, Any]]:
        prolog = self.read_non_data()
        key = self.read_key()
        self._mark_internal_tables((), key[:-1], 'hidden')
        self.match_or_err('\n', 'expecting newline after table key')
        data = self.read_table_assignments(key)

        if key[-1] == -1:
            key = (*key[:-1], id(data))
        self.style.set_value_style(key, data, _ValueStyle(prolog=prolog, display='regular'))
        return key, data

    def read(self) -> Dict[str, Any]:
        prolog = self.read_non_data()
        epilog = ''
        data = self.read_table_assignments(())

        while self.is_not_empty():
            p = self.position
            epilog = self.read_non_data()
            if self.is_not_empty():
                self.position = p
                epilog = ''
                key, value = self.read_regular_table()
                self._enter(data, key, value)
            else:
                break

        self.style.set_value_style((), data, _ValueStyle(prolog=prolog, epilog=epilog, display='regular'))
        return data


def key2path(key: str) -> KEY_T:
    return _TomlParser(key, None).read_key(())


def loads(data: str, file_name: Optional[str] = None) -> Tuple[Dict[str, Any], DUMPS_T]:
    reader = _TomlParser(data, file_name)
    data = reader.read()

    return data, lambda x: _Writer(reader.style).write(x)


def load(file: Union[Path, str]) -> Tuple[Dict[str, Any], DUMPS_T]:
    file = Path(file)
    return loads(file.read_text(), str(file.absolute()))


def dumps(data: Dict[str, Any]) -> str:
    return _Writer().write(data)



from abc import abstractmethod, ABC
from typing import TypeVar, Optional, Callable, Type, Union

_T = TypeVar("_T")
# noinspection PyTypeChecker
_C = TypeVar('_C', bound=Type)


def unone(v: Optional[_T], on_none: Callable[[], _T]) -> _T:
    """
    :param v: the value to check
    :param on_none: callable to execute in the case where v is None
    :return:  `v` if `v` is not None otherwise `on_none()`
    """
    if v is None:
        return on_none()
    return v


def unone_raise(v: Optional[_T], on_none: Callable[[], Exception] = lambda: ValueError('unexpected None')) -> _T:
    """
    :param v: the value to check
    :param on_none: callable to execute in the case where v is None
    :return:  `v` if `v` is not None otherwise raises the exception returned by `on_none()`
    """
    if v is None:
        raise on_none()
    return v


def take_if(value: _T, predicate: Union[Callable[[_T], bool], bool]) -> Optional[_T]:
    """
    :param value: the value to check
    :param predicate: predicate accepting the value or a boolean indicating if accepting the value
    :return: `value` if `predicate(value)` is `True` otherwise `None`
    """

    if callable(predicate):
        if predicate(value):
            return value
    elif predicate:
        return value
    return None


# Common Exceptions

class IllegalStateException(Exception):
    ...


class UnsupportedOperationException(Exception):
    ...


class NoSuchElementException(Exception):
    ...


# Common classes

class Closeable(ABC):
    @abstractmethod
    def close(self): ...

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



from abc import abstractmethod, ABC
from base64 import b64encode
from dataclasses import dataclass
from pathlib import Path
from typing import List, Union, Optional, Tuple, Any, Protocol, Iterable, TYPE_CHECKING

from pkm.api.dependencies.dependency import Dependency
from pkm.api.packages.package import Package
from pkm.api.packages.package_metadata import PackageMetadata
from pkm.api.versions.version_specifiers import AllowAllVersions
from pkm.utils.iterators import partition

if TYPE_CHECKING:
    from pkm.api.environments.environment import Environment
    from pkm.api.repositories.repository_management import RepositoryManagement


class Repository(Protocol):

    @property
    @abstractmethod
    def name(self) -> str:
        ...

    @abstractmethod
    def match(self, dependency: Union[Dependency, str], env: "Environment") -> List[Package]:
        """
        :param dependency: the dependency to match (or a pep508 string representing it)
        :param env: the environment that the returned packages should be compatible with
        :return: list of all the packages in this repository that match the given `dependency`
        """

    def list(self, package_name: str, env: "Environment") -> List[Package]:
        """
        :param package_name: the package to match
        :param env: the environment that the returned packages should be compatible with
        :return: list of all the packages that match the given `package_name`
        """
        return self.match(Dependency(package_name, AllowAllVersions), env)

    @property
    @abstractmethod
    def publisher(self) -> Optional["RepositoryPublisher"]:
        """
        :return: if this repository is 'publishable' returns its publisher
        """

    # noinspection PyMethodMayBeStatic
    def accepted_url_protocols(self) -> Iterable[str]:
        """
        :return: sequence of url-dependency protocols that this repository can handle
        """
        return ()

    # noinspection PyMethodMayBeStatic
    def accept_non_url_packages(self) -> bool:
        """
        :return: True if this repository should be used for non url packages
        """
        return True


# noinspection PyMethodMayBeStatic
class AbstractRepository(Repository, ABC):

    def __init__(self, name: str):
        self._name = name

    @property
    def name(self) -> str:
        return self._name

    def match(self, dependency: Union[Dependency, str], env: "Environment") -> List[Package]:
        if isinstance(dependency, str):
            dependency = Dependency.parse(dependency)

        matched = [d for d in self._do_match(dependency, env) if d.is_compatible_with(env)]
        filtered = self._filter_prereleases(matched, dependency)
        return self._sort_by_priority(dependency, filtered)

    @property
    def publisher(self) -> Optional["RepositoryPublisher"]:
        return None

    def _filter_prereleases(self, packages: List[Package], dependency: Dependency) -> List[Package]:
        if dependency.version_spec.allows_pre_or_dev_releases():
            return packages
        pre_release, rest = partition(packages, lambda it: it.version.is_pre_or_dev_release())
        return rest or packages

    def _sort_by_priority(self, dependency: Dependency, packages: List[Package]) -> List[Package]:
        """
        sorts `matches` by the required priority
        :param dependency: the dependency that resulted in the given `packages`
        :param packages: the packages that were the result `_do_match(dependency)`
        :return: sorted packages by priority (first is more important than last)
        """
        packages.sort(key=lambda it: it.version, reverse=True)
        return packages

    @abstractmethod
    def _do_match(self, dependency: Dependency, env: "Environment") -> List[Package]:
        """
        IMPLEMENTATION NOTICE:
            you don't have to filter pre-releases or packages based on the given environment
            it is handled for you in the `match` method that call this one.

        :param dependency: the dependency to match
        :param env: the environment that the returned packages should be applicable with
        :return: list of all the packages in this repository that match the given `dependency` version spec
        """


class RepositoryPublisher:
    def __init__(self, repository_name: str):
        self.repository_name = repository_name

    # noinspection PyMethodMayBeStatic
    def required_authentication_fields(self) -> List[str]:
        return ['username', 'password']

    @abstractmethod
    def publish(self, auth: "Authentication", package_meta: PackageMetadata, distribution: Path):
        """
        publish a `distribution` belonging to the given `package_meta` into the repository (registering it if needed)
        :param auth: authentication object filled with the fields that were
                     returned by the method `required_authentication_fields`
        :param package_meta: metadata for the package that this distribution belongs to
        :param distribution: the distribution archive (e.g., wheel, sdist)
        """


@dataclass(frozen=True, eq=True)
class Authentication:
    username: str
    password: str

    def as_basic_auth_header(self) -> Tuple[str, str]:
        return 'Authorization', f'Basic {b64encode(f"{self.username}:{self.password}".encode()).decode("ascii")}'


class RepositoryBuilder(ABC):

    def __init__(self, name: str):
        self.name = name

    @abstractmethod
    def build(self, name: str, packages_limit: Optional[List[str]],
              **kwargs: Any) -> Repository:
        """
        build a new repository instance using the given `kwargs`
        :param name: name for the created repository
        :param packages_limit: list of packages the user ask the repository to be limited to
                (or None if no such request was made)
        :param kwargs: arguments for the instance creation, may be defined by derived classes
        :return: the created instance
        """


class HasAttachedRepository(ABC):

    @property
    @abstractmethod
    def repository_management(self) -> "RepositoryManagement":
        ...

    @property
    def attached_repository(self) -> Repository:
        """
        :return: the repository that is attached to this artifact
        """
        return self.repository_management.attached_repo



from __future__ import annotations
import hashlib
import io
import re
from base64 import urlsafe_b64encode
from dataclasses import dataclass
from pathlib import Path
from typing import Protocol, IO, Union, Iterable, TypeVar, Callable, Mapping

from pkm.utils.io_streams import chunks
from pkm.utils.types import SupportHashCode

_T = TypeVar("_T")
_K = TypeVar("_K")
_V = TypeVar("_V")
_H = TypeVar("_H", bound=SupportHashCode)


class HashBuilder:
    def __init__(self):
        self._result = 7

    def regular(self, arg: _H) -> HashBuilder:
        """
        add the regular hash value of `arg` (computed by calling hash(arg))
        :param arg: the argument to add hash for
        :return: self for chaining
        """
        self._result = self._result * 31 + hash(arg)
        return self

    def regulars(self, *args: _H) -> HashBuilder:
        """
        add the regular hash value of each item in `args` (computed by calling hash(arg))
        :param args: the arguments to add hash for
        :return: self for chaining
        """
        return self.ordered_seq(args)

    def ordered_seq(self, seq: Iterable[_T], item_hash: Callable[[_T], int] = hash) -> HashBuilder:
        """
        adds an ordered sequence hash - computes a hash value for the given seq
        :param seq: the sequence to compute hash to
        :param item_hash: a function that can compute hash for the sequence items
        :return: self for chaining
        """
        result = self._result
        for item in seq:
            result = result * 31 + item_hash(item)
        self._result = result

        return self

    def unordered_seq(self, seq: Iterable[_T], item_hash: Callable[[_T], int] = hash) -> HashBuilder:
        """
        adds an unordered sequence hash - computes a hash value for the given seq
        :param seq: the sequence to compute hash to
        :param item_hash: a function that can compute hash for the sequence items
        :return: self for chaining
        """
        result = 7 * 31
        for item in seq:
            result ^= item_hash(item)

        self._result = self._result * 31 + result
        return self

    def unordered_mapping(
            self, m: Mapping[_K, _V], key_hash: Callable[[_K], int] = hash,
            value_hash: Callable[[_V], int] = hash) -> HashBuilder:
        """
        adds unordered maping hash - computes a hash value for the given dict, the hash will be the same for all dicts
        with the same elements without considering their insertion order
        :param m: the mapping to compute hash to
        :param key_hash: a function that can compute hash for the dict keys
        :param value_hash: a function that can compute hash for the dict values
        :return: self for chaining
        """
        result = 0
        for k, v in m.items():
            kvh = (7 * 31 + key_hash(k)) * 31 + value_hash(v)
            result ^= kvh

        self._result = self._result * 31 + result
        return self

    def build(self) -> int:
        """
        :return: the built hash
        """
        return self._result


_SIG_DELIM_RX = re.compile("\\s*=\\s*")


class HashDigester(Protocol):
    def digest(self) -> bytes:
        ...

    def hexdigest(self) -> str:
        ...

    def update(self, __data: Union[bytes, bytearray, memoryview]) -> None:
        ...


def stream(hd: HashDigester, file: Union[IO, Path], chunk_size: int = io.DEFAULT_BUFFER_SIZE):
    """
    streams a file into `hd` (via `hd.update(chunk)`) using chunks of the given `chunk_size`
    :param hd: the hash digester to stream into
    :param file: the file to stream
    :param chunk_size: the size of the chunk to use, defaults to `io.DEFAULT_BUFFER_SIZE`
    """
    if isinstance(file, Path):
        with file.open('rb') as source_fd:
            stream(hd, source_fd, chunk_size)
    else:
        for chunk in chunks(file, chunk_size):
            hd.update(chunk)


@dataclass
class HashSignature:
    hash_type: str
    hash_value: str

    def _encode_hash(self, hashd: HashDigester) -> str:
        return type(self).encode_hash(hashd)

    def validate_against(self, file: Path) -> bool:
        if not hasattr(hashlib, self.hash_type):
            raise KeyError(f"Cannot validate archive, Unsupported Hash {self.hash_type}")

        hash_computer = getattr(hashlib, self.hash_type)()
        stream(hash_computer, file)
        computed_hash = self._encode_hash(hash_computer)
        return computed_hash == self.hash_value

    def __str__(self):
        return f"{self.hash_type}={self.hash_value}"

    def __repr__(self):
        return f"HashSignature({self})"

    @classmethod
    def encode_hash(cls, hashd: HashDigester):
        return hashd.hexdigest()

    @classmethod
    def create_hex_encoded(cls, hash_type: str, hash_value: str):
        return HashSignature(hash_type, hash_value)

    @classmethod
    def compute_hex_encoded(cls, hash_function: str, file: Path) -> HashSignature:
        hashd = getattr(hashlib, hash_function)()
        stream(hashd, file)
        encoded = HashSignature.encode_hash(hashd)
        return HashSignature(hash_function, encoded)

    @classmethod
    def parse_hex_encoded(cls, signature: str) -> HashSignature:
        parts = _SIG_DELIM_RX.split(signature)
        if len(parts) != 2:
            raise ValueError('unsupported signature, expecting format <hash_type>=<hash_value>')

        return HashSignature(*parts)

    @classmethod
    def parse_urlsafe_base64_nopad_encoded(cls, signature: str) -> HashSignature:
        parts = _SIG_DELIM_RX.split(signature)
        if len(parts) != 2:
            raise ValueError(f'unsupported signature, expecting format <hash_type>=<hash_value>, got: {signature}')

        return _UrlsafeBase64NopadHashSignature(*parts)

    @classmethod
    def compute_urlsafe_base64_nopad_encoded(cls, hash_function: str, file: Path) -> HashSignature:
        hashd = getattr(hashlib, hash_function)()
        stream(hashd, file)
        encoded = _UrlsafeBase64NopadHashSignature.encode_hash(hashd)
        return _UrlsafeBase64NopadHashSignature(hash_function, encoded)

    @classmethod
    def create_urlsafe_base64_nopad_encoded(cls, hash_function: str, digest: HashDigester) -> HashSignature:
        return _UrlsafeBase64NopadHashSignature(hash_function, _UrlsafeBase64NopadHashSignature.encode_hash(digest))


class _UrlsafeBase64NopadHashSignature(HashSignature):

    @classmethod
    def encode_hash(cls, hashd: HashDigester):
        return urlsafe_b64encode(hashd.digest()).decode("latin1").rstrip("=")



from abc import abstractmethod
from datetime import datetime
from typing import Dict

from pkm.config.configuration import TomlFileConfiguration
from typing import Protocol
from http.client import HTTPResponse

from pkm.utils.functionals import run_once

TF_HTTP = '%a, %d %b %Y %H:%M:%S GMT'
TF_PACKAGE_UPLOAD_TIME = '%Y-%m-%dT%H:%M:%S'


class CacheDirective(Protocol):

    @abstractmethod
    def is_cache_valid(self, fetch_info: TomlFileConfiguration) -> bool:
        ...

    def add_headers(self, fetch_info: TomlFileConfiguration, headers: Dict[str, str]):
        ...

    def store_result(self, resp: HTTPResponse) -> bool:
        return True

    @classmethod
    @run_once
    def allways(cls) -> "CacheDirective":
        class _Allways(CacheDirective):
            def is_cache_valid(self, fetch_info: TomlFileConfiguration) -> bool:
                return True

        return _Allways()

    @classmethod
    @run_once
    def ask_for_update(cls):
        class _AskUpdate(CacheDirective):
            def is_cache_valid(self, fetch_info: TomlFileConfiguration) -> bool:
                return False

            def add_headers(self, fetch_info: TomlFileConfiguration, headers: Dict[str, str]):
                etag = fetch_info['etag']
                if etag:
                    headers['If-None-Match'] = etag
                else:
                    fetch_time = fetch_info['fetch-time']
                    if fetch_time is not None:
                        headers['If-Modified-Since'] = fetch_time

        return _AskUpdate()

    @classmethod
    def if_retrieved_after(cls, dt: datetime):
        class _After(CacheDirective):
            def is_cache_valid(self, fetch_info: TomlFileConfiguration) -> bool:
                fetch_time = fetch_info['fetch-time']
                if fetch_time is None:
                    return False

                fetch_time = datetime.strptime(fetch_time, TF_HTTP)
                return fetch_time >= dt

        return _After()

    @classmethod
    @run_once
    def never(cls) -> "CacheDirective":
        class _Never(CacheDirective):
            def is_cache_valid(self, fetch_info: TomlFileConfiguration) -> bool:
                return False

            def store_result(self, resp: HTTPResponse) -> bool:
                return False

        return _Never()



import os
from contextlib import contextmanager
from copy import copy, deepcopy
from dataclasses import replace
from pathlib import Path
from textwrap import dedent
from typing import List, Union, Optional, ContextManager, Dict

from pkm.api.dependencies.dependency import Dependency
from pkm.api.distributions.distinfo import DistInfo, PackageInstallationInfo
from pkm.api.distributions.wheel_distribution import WheelDistribution
from pkm.api.packages.package import PackageDescriptor
from pkm.api.packages.package_installation import PackageInstallationTarget
from pkm.api.packages.site_packages import InstalledPackage
from pkm.api.projects.project import Project
from pkm.api.projects.pyproject_configuration import PyProjectConfiguration, PkmApplicationConfig, ProjectConfig, \
    BuildSystemConfig
from pkm.api.repositories.repositories_configuration import RepositoriesConfiguration, RepositoryInstanceConfig
from pkm.api.repositories.repository_loader import REPOSITORIES_CONFIGURATION_PATH
from pkm.distributions.executables import Executables
from pkm.pep517_builders.pkm_builders import build_wheel
from pkm.repositories.local_packages_repository import LOCAL_PACKAGES_REPOSITORY_TYPE
from pkm.utils.entrypoints import EntryPoint, ObjectReference
from pkm.utils.files import temp_dir, mkdir, CopyTransaction, path_to
from pkm.utils.hashes import HashSignature
from pkm.utils.properties import cached_property

_CONTAINERIZED_WRAPPER_SUFFIX = "_containerized"
_CONTAINERIZED_WRAPPER_VERSION = "0.0.1"
_CONTAINERIZED_TARGET_USED_PATHS = ["purelib", "data", "scripts"]

CONTAINERIZED_APP_DATA_PATH = "__container__"
CONTAINERIZED_APP_SITE_PATH = "__container__/site"
CONTAINERIZED_APP_BIN_PATH = "__container__/bin"


class ContainerizedApplication:
    def __init__(self, package: InstalledPackage, target: PackageInstallationTarget):
        self.package = package
        self._target = target

    @cached_property
    def app_package(self) -> InstalledPackage:
        site = self._target.site_packages
        result = None
        if self.package.name.endswith(_CONTAINERIZED_WRAPPER_SUFFIX):
            result = site.installed_package(self.package.name[:-len(_CONTAINERIZED_WRAPPER_SUFFIX)])

        return result or site.installed_package(self.package.name)

    @property
    def installation_target(self) -> PackageInstallationTarget:
        return self._target

    def uninstall(self):
        self.package.uninstall()

    def list_installed_plugins(self) -> List[InstalledPackage]:
        return [it
                for it in self._target.site_packages.installed_packages()
                if it.dist_info.is_user_requested() and it.name != self.package.name]


class ContainerizedApplications:
    def __init__(self, target: PackageInstallationTarget):
        self._target = target

    def containers(self) -> List[ContainerizedApplication]:
        return [
            ContainerizedApplication(it, self._target_of(it.name))
            for it in self._target.site_packages.installed_packages()
            if it.dist_info.is_app_container()]

    def _target_of(self, package: str) -> PackageInstallationTarget:
        app_base = Path(self._target.purelib) / PackageDescriptor.normalize_src_package_name(package)
        app_site = str(app_base / CONTAINERIZED_APP_SITE_PATH)
        app_bin = str(app_base / CONTAINERIZED_APP_BIN_PATH)
        app_data = str(app_base / CONTAINERIZED_APP_DATA_PATH)

        return replace(self._target, platlib=app_site, purelib=app_site, scripts=app_bin, data=app_data)

    def container_of(self, package: str) -> Optional[ContainerizedApplication]:
        spacks = self._target.site_packages
        package_cnt = package + _CONTAINERIZED_WRAPPER_SUFFIX
        if (cnt := spacks.installed_package(package)) or (cnt := spacks.installed_package(package_cnt)):
            if cnt.dist_info.is_app_container():
                return ContainerizedApplication(cnt, self._target_of(cnt.name))
            raise ValueError(f"installed package: {cnt.descriptor} is not containerized")
        return None

    @contextmanager
    def _wrapper_project(self, app: Union[Dependency, "Project"]) -> ContextManager["Project"]:
        from pkm.api.projects.project import Project
        with temp_dir() as tdir:
            pyprj = PyProjectConfiguration.load(tdir / 'pyproject.toml')

            if isinstance(app, Project):
                repocfg = RepositoriesConfiguration.load(tdir / REPOSITORIES_CONFIGURATION_PATH)
                repocfg.repositories = [RepositoryInstanceConfig.from_config('wrapped-project', {
                    'type': LOCAL_PACKAGES_REPOSITORY_TYPE,
                    'projects': [app.path]
                })]

                repocfg.save()
                app = app.descriptor.to_dependency()

            pyprj.project = ProjectConfig.from_config({
                'name': app.package_name + _CONTAINERIZED_WRAPPER_SUFFIX,
                'version': _CONTAINERIZED_WRAPPER_VERSION,
            })

            pyprj.build_system = BuildSystemConfig([], "pkm_containerized", ["."])

            pyprj.pkm_application = PkmApplicationConfig(True, [app], {}, [app.package_name])
            pyprj.save()
            yield Project.load(tdir)

    def _install(
            self, app: "Project", editable: bool = True) -> ContainerizedApplication:

        contained_target = self._target_of(app.name)
        app_dir = Path(contained_target.purelib).parent.parent

        with temp_dir() as tdir, CopyTransaction() as ct:

            if Path(contained_target.purelib).exists():
                contained_target.force_remove(app.name)
            else:
                for path in _CONTAINERIZED_TARGET_USED_PATHS:
                    mkdir(Path(getattr(contained_target, path)))

            contained_site = contained_target.site_packages
            app_config = app.config.pkm_application

            contained_target.install(
                app_config.dependencies, dependencies_override=app_config.dependency_overwrites,
                repository=app.attached_repository,
                editables={d.package_name: editable for d in app_config.dependencies},
                updates=[d.package_name for d in app_config.dependencies])

            actual_app = Project(deepcopy(app.config))
            actual_app.config.project.dependencies = app_config.dependencies
            actual_app.config.pkm_application = None
            WheelDistribution(app.descriptor, build_wheel(actual_app, tdir / "whl", editable=editable)) \
                .install_to(contained_target, app.descriptor.to_dependency(),
                            PackageInstallationInfo(editable=editable))

            contained_site.reload()

            for path in _CONTAINERIZED_TARGET_USED_PATHS:
                ct.touch_tree(Path(getattr(contained_target, path)))

            contained_app = contained_site.installed_package(app.name)
            contained_distinfo = contained_app.dist_info

            # now create the dist-info
            dist_info_path = app_dir.parent / contained_distinfo.path.name
            if dist_info_path.exists():
                ct.rm(dist_info_path)
            dist_info = DistInfo.load(mkdir(dist_info_path))

            ct.copy_tree(contained_distinfo.path, dist_info.path)
            ct.touch(app_dir / "__init__.py", True)

            # collect script entrypoints
            script_entrypoints: List[EntryPoint] = []
            apps_to_expose = app_config.exposed_packages + [app.name]
            for exposed_app in apps_to_expose:
                if installed := contained_site.installed_package(exposed_app):
                    script_entrypoints.extend(
                        it for it in installed.dist_info.load_entrypoints_cfg().entrypoints if it.is_script())

            # submit script entrypoints
            app_entrypoints = dist_info.load_entrypoints_cfg()
            entrypoints = []
            if script_entrypoints:
                app_entrypoints_script = (app_dir / "entrypoints.py")
                app_entrypoints_script.write_text(_entrypoints_script(script_entrypoints))
                ct.touch(app_entrypoints_script)
                scripts_path = Path(self._target.scripts)
                for script_ep in script_entrypoints:
                    epn = PackageDescriptor.normalize_src_package_name(script_ep.name)
                    entrypoints.append(
                        script_ep := replace(script_ep, ref=ObjectReference.parse(f"{app_dir.name}.entrypoints:{epn}")))
                    ct.touch(Executables.generate_for_entrypoint(self._target.env, script_ep, scripts_path))

            app_entrypoints.entrypoints = entrypoints
            app_entrypoints.save()

            dist_info.save_installation_info(PackageInstallationInfo(containerized=True, editable=False))
            ct.touch(dist_info.installation_info_path())
            ct.touch(dist_info.user_requested_path(), True)  # mark as user requested

            records = dist_info.load_record_cfg()
            records.records.clear()

            # reuse precomputed hashes in order to sign our records
            # TODO: notify a monitor that we are spending time on signing the package
            precomputed_hashes: Dict[str, HashSignature] = {}
            installation_site = dist_info.path.parent
            for pk in contained_site.installed_packages():
                for record in pk.dist_info.load_record_cfg().records:
                    path_relative_to_site = str(path_to(installation_site, record.absolute_path(pk.dist_info)))
                    precomputed_hashes[path_relative_to_site] = record.hash_signature

            records.sign_files(ct.copied_files, installation_site, precomputed_hashes)
            records.save()

        self._target.site_packages.reload()
        return self.container_of(app.name)

    def _get_or_install(self, app: Union[Dependency, Project], editable: bool = True) -> ContainerizedApplication:
        dep = app
        if isinstance(dep, Project):
            dep = app.descriptor.to_dependency()

        if (container := self.container_of(dep.package_name)) \
                and dep.version_spec.allows_version(container.app_package.version):
            return container

        return self.install(app, editable, True)

    def install(
            self, app: Union[Dependency, Project], editable: bool = True, update: bool = False
    ) -> ContainerizedApplication:
        """
        installs the given app in its own containerized environment
        :param app: the app to install, if not referring to a containerized project will create containerized wrapper
        :param editable: if true, the installation of `app` inside the container will be in editable mode
        :param update: if true, will force re-installation,
            even if the same version is already installed in the relevant site
        :return: containerized application controller class for the performed installation
        """

        if not update:
            return self._get_or_install(app, editable)

        if isinstance(app, Project) and app.is_containerized_application():
            return self._install(app, editable)

        with self._wrapper_project(app) as prj:
            return self.install(prj, editable, True)


def _entrypoints_script(epoints: List[EntryPoint]):
    def define_epfunc(ep: EntryPoint):
        epn = PackageDescriptor.normalize_src_package_name(ep.name)
        ref = ep.ref
        return f"def {epn}():del sys.modules[__package__];" \
               f"import {ref.module_path};" \
               f"{ref.module_path}.{ref.object_path}()"

    return dedent(f"""

    # setting up the application container
    import site 
    import sys
    import os
    from pathlib import Path

    container_path = Path(__file__).parent
    site.addsitedir(container_path / '__container__/site')
    sys.path.insert(1, sys.path.pop())

    old_path = os.environ['PATH']
    new_path = str(container_path / '{CONTAINERIZED_APP_BIN_PATH}')
    if old_path:
        new_path = old_path + os.pathsep + new_path
    os.environ['PATH'] = new_path
    
    sys.prefix = str(container_path/'{CONTAINERIZED_APP_DATA_PATH}')

    # defining entry points 
    """) + os.linesep.join([define_epfunc(ep) for ep in epoints])



from abc import abstractmethod
from pathlib import Path
from typing import Protocol, Optional, TYPE_CHECKING, List, Dict

from pkm.api.distributions.distinfo import PackageInstallationInfo
from pkm.api.packages.package import PackageDescriptor, Package
from pkm.api.dependencies.dependency import Dependency
from pkm.api.packages.package_metadata import PackageMetadata

if TYPE_CHECKING:
    from pkm.api.environments.environment import Environment
    from pkm.api.packages.package import PackageInstallationTarget


class Distribution(Protocol):

    @property
    @abstractmethod
    def owner_package(self) -> PackageDescriptor:
        """
        :return: the package descriptor that this distribution belongs to
        """

    @abstractmethod
    def install_to(self, target: "PackageInstallationTarget", user_request: Optional[Dependency] = None,
                   installation_mode: Optional[PackageInstallationInfo] = None):
        """
        installs this package into the given `env`
        :param target: information about the target to install this distribution into
        :param user_request: if this package was requested by the user,
               supplying this field will mark the installation as user request and save the given info
        :param installation_mode: information about the installation mode to save while installing
        """

    @abstractmethod
    def extract_metadata(self, env: "Environment") -> PackageMetadata:
        """
        extracts and returns metadata from this distribution
        :param env: the environment that this metadata should be relevant to
        :return: the extracted metadata
        """

    @classmethod
    def package_from(cls, distribution: Path, desc: Optional[PackageDescriptor] = None) -> Package:
        """
        creates a package from the given distribution, assumes proper naming conventions for distribution file name
        :param distribution: the path to the distribution
        :param desc: if given, will be used as the descriptor of the package, otherwise a descriptor will be guessed
            from the file naming conventions
        :return: package that upon install will install the given distribution
        """

        from pkm.api.distributions.source_distribution import SourceDistribution
        from pkm.api.distributions.wheel_distribution import WheelDistribution

        desc = desc or PackageDescriptor.from_dist_name(distribution.name)

        if distribution.name.endswith(".whl"):
            return _DistributionPackage(WheelDistribution(desc, distribution))

        return _DistributionPackage(SourceDistribution(desc, distribution))


class _DistributionPackage(Package):

    def __init__(self, dist: Distribution):
        self._dist = dist
        self._env_hash_to_metadata: Dict[str, PackageMetadata] = {}

    @property
    def descriptor(self) -> PackageDescriptor:
        return self._dist.owner_package

    def _metadata(self, env: "Environment"):
        env_hash = env.markers_hash
        if not (meta := self._env_hash_to_metadata.get(env_hash)):
            meta = self._dist.extract_metadata(env)
            self._env_hash_to_metadata[env_hash] = meta
        return meta

    def dependencies(
            self, environment: "Environment",
            extras: Optional[List[str]] = None) -> List["Dependency"]:
        return [d for d in self.published_metadata.dependencies if d.is_applicable_for(environment, extras)]

    def is_compatible_with(self, env: "Environment") -> bool:
        from pkm.api.distributions.wheel_distribution import WheelDistribution

        if isinstance(self._dist, WheelDistribution):
            return env.compatibility_tag_score(self._dist.compute_compatibility_tags()) is not None
        return True

    def install_to(
            self, target: "PackageInstallationTarget", user_request: Optional["Dependency"] = None,
            editable: bool = False):
        self._dist.install_to(target, user_request, PackageInstallationInfo(editable=editable))



from html.parser import HTMLParser
from pathlib import Path
from typing import List, Union, Tuple, Dict, Optional, Callable, Any

from pkm.api.dependencies.dependency import Dependency
from pkm.api.environments.environment import Environment
from pkm.api.packages.package import Package, PackageDescriptor
from pkm.api.packages.standard_package import PackageArtifact, AbstractPackage
from pkm.api.repositories.repository import Repository, RepositoryBuilder, AbstractRepository
from pkm.api.versions.version import Version
from pkm.api.versions.version_specifiers import VersionSpecifier
from pkm.utils.http.cache_directive import CacheDirective
from pkm.utils.http.http_client import HttpClient, Url
from pkm.utils.iterators import groupby
from pkm.utils.strings import endswith_any, without_suffix


class SimpleRepository(AbstractRepository):
    """
    implementation of pep503 simple repository
    """

    def __init__(self, name: str, http_client: HttpClient, url: str):
        super().__init__(name)
        self._http_client = http_client
        self._url = url
        self._base_url = Url.parse(url).connection_part()
        self._packages: Dict[str, Dict[str, Package]] = {}  # name -> version -> package

    def _do_match(self, dependency: Dependency, env: Environment) -> List[Package]:
        # monitor.on_dependency_match(dependency)
        if not (version_to_package := self._packages.get(dependency.package_name)):
            data = self._http_client.fetch_resource(f"{self._url}/{dependency.package_name}").data
            extractor = _HtmlArtifactsExtractor(self._base_url)
            extractor.feed(data.read_text())

            all_artifacts = extractor.artifacts
            grouped_by_version: Dict[str, List[PackageArtifact]] = groupby(
                all_artifacts, lambda a: _extract_version(a.file_name))

            version_to_package = {
                version_str: _SimplePackage(
                    self._http_client,
                    PackageDescriptor(dependency.package_name, Version.parse(version_str)),
                    version_artifacts)

                for version_str, version_artifacts in grouped_by_version.items()
            }
            self._packages[dependency.package_name] = version_to_package

        return [p for p in version_to_package.values() if dependency.version_spec.allows_version(p.version)]


_DISTRIBUTION_EXTENSIONS = (".whl", ".tar.gz", ".zip")


def _extract_version(filename: str) -> str:
    result = without_suffix(filename, suffix := endswith_any(filename, _DISTRIBUTION_EXTENSIONS))
    if suffix == '.whl':
        return result.split("-")[1]
    else:
        return result.split("-")[-1]


class _HtmlArtifactsExtractor(HTMLParser):

    def __init__(self, base_url: str, *, convert_charrefs: bool = ...) -> None:
        super().__init__(convert_charrefs=convert_charrefs)
        self._text_handler: Optional[Callable[[str], None]] = None
        self.artifacts: List[PackageArtifact] = []
        self._base_url = base_url

    def handle_data(self, data: str) -> None:
        if self._text_handler:
            self._text_handler(data)

    def handle_endtag(self, tag: str) -> None:
        if self._text_handler:
            self._text_handler('')

    def handle_starttag(self, tag: str, attrs: List[Tuple[str, Union[str, None]]]) -> None:
        self._text_handler = \
            lambda txt: self.handle_element(tag, {kv[0]: kv[1] for kv in attrs if kv[1] is not None}, txt)

    def handle_element(self, tag: str, attrs: Dict[str, str], text: str):
        self._text_handler = None
        if tag == 'a' and 'href' in attrs:
            requires_python = None
            if requires_python_str := attrs.get('data-requires-python'):
                requires_python = VersionSpecifier.parse(requires_python_str)

            url = attrs.get('href')
            if not url.startswith('http'):
                url = f"{self._base_url}/{url.lstrip('/')}"

            if endswith_any(text, _DISTRIBUTION_EXTENSIONS):
                self.artifacts.append(PackageArtifact(text, requires_python, {'url': url}))


class _SimplePackage(AbstractPackage):
    def __init__(self, http_client: HttpClient, descriptor: PackageDescriptor,
                 artifacts: List[PackageArtifact]):
        super().__init__(descriptor, artifacts)
        self._http_client = http_client

    def _retrieve_artifact(self, artifact: PackageArtifact) -> Path:
        return self._http_client.fetch_resource(
            artifact.other_info['url'], CacheDirective.allways(),
            resource_name=f"{self.name} {self.version}"
        ).data


class SimpleRepositoryBuilder(RepositoryBuilder):
    def __init__(self, http_client: HttpClient):
        super().__init__("simple")
        self._http_client = http_client

    def build(self, name: Optional[str], packages_limit: Optional[List[str]], **kwargs: Any) -> Repository:
        if not (url := kwargs.get('url')):
            raise KeyError("url field is required to build 'simple' repository")
        return SimpleRepository(name or url, self._http_client, str(url).rstrip('/'))



import os.path
import platform
import re
import subprocess
from pathlib import Path
from subprocess import CalledProcessError
from typing import List, Optional, Set, Union

from pkm.api.dependencies.dependency import Dependency
from pkm.api.environments.environment import Environment
from pkm.api.environments.environment_builder import EnvironmentBuilder
from pkm.api.packages.package import PackageDescriptor, Package
from pkm.api.packages.package_installation import PackageInstallationTarget
from pkm.api.repositories.repository import AbstractRepository
from pkm.api.versions.version import Version
from pkm.utils.properties import cached_property
from pkm.utils.systems import is_executable

_DEFAULT_PKG_EXTRAS = {'pip', 'wheel', 'setuptools'}


class InstalledPythonsRepository(AbstractRepository):

    def __init__(self):
        super().__init__('local-pythons')

    def list(self, package_name: str = 'python', env: Environment = None) -> List[Package]:
        return super().list(package_name, env)

    @cached_property
    def _interpreters(self) -> List["LocalInterpreterPackage"]:
        result: List[LocalInterpreterPackage] = []
        interpreters_in_path = _interpreters_in_path()
        executeables_matched: Set[Path] = set()
        for interpreter_path in interpreters_in_path:
            try:
                cmdout = subprocess.run(
                    [str(interpreter_path), "-c",
                     "import platform;import sys;print(platform.python_version());print(sys.executable)"],
                    capture_output=True)
                cmdout.check_returncode()

                version_str, executable = cmdout.stdout.decode().strip().splitlines(keepends=False)
                executable = Path(executable.strip()).resolve()

                if executable in executeables_matched:
                    continue

                executeables_matched.add(executable)

                result.append(LocalInterpreterPackage(
                    executable,
                    PackageDescriptor("python", Version.parse(version_str.strip())),
                    _DEFAULT_PKG_EXTRAS))

            except (ChildProcessError, CalledProcessError):
                # import traceback
                # traceback.print_exc()
                pass  # skip this interpreter

        return sorted(result, key=lambda p: p.version, reverse=True)

    def _do_match(self, dependency: Dependency, env: Environment) -> List[Package]:
        # monitor.on_dependency_match(dependency)
        extras = set(dependency.extras) if dependency.extras is not None else _DEFAULT_PKG_EXTRAS

        return [
            p.with_extras(extras)
            for p in self._interpreters
            if dependency.version_spec.allows_version(p.version)]


class LocalInterpreterPackage(Package):

    def __init__(self, interpreter: Path, desc: PackageDescriptor, extras: Set[str]):
        self._interpreter = interpreter
        self._desc = desc
        self._extras = extras

    def dependencies(self, environment: "Environment", extras: Optional[List[str]] = None) -> List["Dependency"]:
        return []

    def with_extras(self, extras: Set[str]) -> "LocalInterpreterPackage":
        if self._extras == extras:
            return self
        return LocalInterpreterPackage(self._interpreter, self._desc, extras)

    @property
    def descriptor(self) -> PackageDescriptor:
        return self._desc

    def is_compatible_with(self, env: Environment):
        return True

    def to_environment(self) -> Environment:
        return Environment(env_path=self._interpreter.parent, interpreter_path=self._interpreter)

    def install_to(
            self, target: Union["PackageInstallationTarget", Environment], user_request: Optional["Dependency"] = None,
            editable: bool = False):
        env_dir = target.path if isinstance(target, Environment) else target.env.path
        EnvironmentBuilder.create(env_dir, self._interpreter.absolute())


_OS = platform.system()
_PYTHON_EXEC_RX = re.compile(r"python-?[\d.]*(.exe)?")


def _interpreters_in_path() -> Set[Path]:
    path_parts = [path for it in (os.environ.get("PATH") or "").split(os.pathsep) if (path := Path(it)).exists()]

    def is_python_executeable(file: Path):
        return _PYTHON_EXEC_RX.fullmatch(file.name.lower()) and is_executable(file)

    return {
        file.resolve()
        for path in path_parts
        for file in path.iterdir()
        if is_python_executeable(file)}



import typing
from typing import Any, TypeVar, Protocol, Generic, Iterator

import sys

_T = TypeVar("_T")


class IterableWithLen(Generic[_T]):
    def __len__(self) -> int:
        ...

    def __iter__(self) -> Iterator[_T]:
        ...


class SupportsLessThan(Protocol):
    def __lt__(self, __other: Any) -> bool: ...


class SupportsLessThanEq(Protocol):
    def __lt__(self, __other: Any) -> bool: ...

    def __eq__(self, other: Any) -> bool: ...


class SupportHashCode(Protocol):
    def __hash__(self): ...

    def __eq__(self, other): ...


# Credits: some of the following type helpers are taken directly from typing_extensions
class _Immutable:
    """Mixin to indicate that object should not be copied."""
    __slots__ = ()

    def __copy__(self):
        return self

    def __deepcopy__(self, memo):
        return self


class ParamSpecArgs(_Immutable):
    """The args for a ParamSpec object.

    Given a ParamSpec object P, P.args is an instance of ParamSpecArgs.

    ParamSpecArgs objects have a reference back to their ParamSpec:

    P.args.__origin__ is P

    This type is meant for runtime introspection and has no special meaning to
    static type checkers.
    """

    def __init__(self, origin):
        self.__origin__ = origin

    def __repr__(self):
        return f"{self.__origin__.__name__}.args"


class ParamSpecKwargs(_Immutable):
    """The kwargs for a ParamSpec object.

    Given a ParamSpec object P, P.kwargs is an instance of ParamSpecKwargs.

    ParamSpecKwargs objects have a reference back to their ParamSpec:

    P.kwargs.__origin__ is P

    This type is meant for runtime introspection and has no special meaning to
    static type checkers.
    """

    def __init__(self, origin):
        self.__origin__ = origin

    def __repr__(self):
        return f"{self.__origin__.__name__}.kwargs"


# Inherits from list as a workaround for Callable checks in Python < 3.9.2.
class ParamSpec(list):
    """Parameter specification variable.

    Usage::

       P = ParamSpec('P')

    Parameter specification variables exist primarily for the benefit of static
    type checkers.  They are used to forward the parameter types of one
    callable to another callable, a pattern commonly found in higher order
    functions and decorators.  They are only valid when used in ``Concatenate``,
    or s the first argument to ``Callable``. In Python 3.10 and higher,
    they are also supported in user-defined Generics at runtime.
    See class Generic for more information on generic types.  An
    example for annotating a decorator::

       T = TypeVar('T')
       P = ParamSpec('P')

       def add_logging(f: Callable[P, T]) -> Callable[P, T]:
           '''A type-safe decorator to add logging to a function.'''
           def inner(*args: P.args, **kwargs: P.kwargs) -> T:
               logging.info(f'{f.__name__} was called')
               return f(*args, **kwargs)
           return inner

       @add_logging
       def add_two(x: float, y: float) -> float:
           '''Add two numbers together.'''
           return x + y

    Parameter specification variables defined with covariant=True or
    contravariant=True can be used to declare covariant or contravariant
    generic types.  These keyword arguments are valid, but their actual semantics
    are yet to be decided.  See PEP 612 for details.

    Parameter specification variables can be introspected. e.g.:

       P.__name__ == 'T'
       P.__bound__ == None
       P.__covariant__ == False
       P.__contravariant__ == False

    Note that only parameter specification variables defined in global scope can
    be pickled.
    """

    # Trick Generic __parameters__.
    __class__ = TypeVar

    @property
    def args(self):
        return ParamSpecArgs(self)

    @property
    def kwargs(self):
        return ParamSpecKwargs(self)

    def __init__(self, name, *, bound=None, covariant=False, contravariant=False):
        super().__init__([self])
        self.__name__ = name
        self.__covariant__ = bool(covariant)
        self.__contravariant__ = bool(contravariant)
        if bound:
            self.__bound__ = typing._type_check(bound, 'Bound must be a type.')  # noqa
        else:
            self.__bound__ = None

        # for pickling:
        try:
            def_mod = sys._getframe(1).f_globals.get('__name__', '__main__')  # noqa
        except (AttributeError, ValueError):
            def_mod = None
        if def_mod != 'typing_extensions':
            self.__module__ = def_mod

    def __repr__(self):
        if self.__covariant__:
            prefix = '+'
        elif self.__contravariant__:
            prefix = '-'
        else:
            prefix = '~'
        return prefix + self.__name__

    def __hash__(self):
        return object.__hash__(self)

    def __eq__(self, other):
        return self is other

    def __reduce__(self):
        return self.__name__

    # Hack to get typing._type_check to pass.
    def __call__(self, *args, **kwargs):
        pass



import functools
from typing import TypeVar, Callable

_T = TypeVar("_T")
_FUNC_T = TypeVar("_FUNC_T", bound=Callable)


def run_once(func: _FUNC_T) -> _FUNC_T:
    computed = False
    result = None

    @functools.wraps(func)
    def func_wrapper(*args, **kwargs):
        nonlocal result
        nonlocal computed

        if not computed:
            result = func(*args, **kwargs)
            computed = True

        return result

    return func_wrapper


def identity(x: _T) -> _T:
    return x

